"""Node domain - everything related to node development and execution"""

from noxus_sdk.nodes.base import BaseNode, NodeConfiguration
from noxus_sdk.nodes.connector import Connector, DataContainer
from noxus_sdk.nodes.schemas import (
    ConfigResponse,
    ExecutionResponse,
    NodeDefinition,
    NodeInput,
    NodeOutput,
)
from noxus_sdk.nodes.types import DataType, NodeCategory, TypeDefinition
from noxus_sdk.nodes.validation import validate_node

__all__ = [
    "BaseNode",
    "ConfigResponse",
    "Connector",
    "DataContainer",
    "DataType",
    "ExecutionResponse",
    "NodeCategory",
    "NodeConfiguration",
    "NodeDefinition",
    "NodeInput",
    "NodeOutput",
    "TypeDefinition",
    "validate_node",
]



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal

import httpx
from loguru import logger
from pydantic import BaseModel

from noxus_sdk.integrations.base import BaseIntegration

if TYPE_CHECKING:
    from uuid import UUID

    from noxus_sdk.plugins.context import RemoteExecutionContext


class NangoProviderCredentials(BaseModel):
    """Configuration for Nango provider credentials."""


class NangoProviderOAuthCredentials(NangoProviderCredentials):
    """Configuration for Nango OAuth credentials.

    Reference:
    https://docs.nango.dev/reference/api/integration/create#option-1
    """

    type: Literal["OAUTH1", "OAUTH2", "TBA"]
    client_id: str
    client_secret: str
    scopes: str = ""
    webhook_secret: str | None = None


class NangoIntegration(BaseIntegration, ABC):
    """Integration created using Nango"""

    type = "nango"

    provider: str
    window_height: int = 600
    window_width: int = 600

    def __init_subclass__(cls) -> None:
        """Set the configuration class for the plugin when the sublcass is created"""

        if not hasattr(cls, "provider"):
            raise ValueError("Provider must be set for Nango integration")

        cls.properties = {
            "provider": cls.provider,
            "window_height": cls.window_height,
            "window_width": cls.window_width,
        }

    @classmethod
    def get_trigger_endpoint(cls, group_id: UUID) -> str | None:
        """Get the trigger endpoint of the integration"""
        return f"/groups/{group_id}/integrations/nango/{cls.provider}"

    @classmethod
    def get_delete_endpoint(cls, group_id: UUID) -> str | None:
        """Get the delete endpoint of the integration"""
        return f"/groups/{group_id}/integrations/nango/{cls.provider}"

    @classmethod
    def get_confirm_endpoint(cls, group_id: UUID) -> str | None:
        """Get the confirm endpoint of the integration"""
        return f"/groups/{group_id}/integrations/nango"

    @classmethod
    @abstractmethod
    def get_provider_credentials(
        cls,
        ctx: RemoteExecutionContext,
    ) -> NangoProviderCredentials:
        """Get the Nango credentials for this integration."""
        raise NotImplementedError(
            "Integration must implement get_provider_credentials() method",
        )

    @classmethod
    async def get_config(cls, ctx: RemoteExecutionContext) -> dict:
        """Get the config of the integration"""
        credentials = cls.get_provider_credentials(ctx)
        return {
            "credentials": credentials.model_dump(),
        }

    @classmethod
    async def get_credentials(cls, data: dict) -> dict:
        """Get the credentials of the integration"""

        connection_id = data["connection_id"]

        nango_secret_key = os.getenv("NANGO_SECRET_KEY")
        if not nango_secret_key:
            raise ValueError("NANGO_SECRET_KEY is not set")

        async with httpx.AsyncClient() as client:
            resp = await client.get(
                f"https://api.nango.dev/connection/{connection_id}",
                headers={"Authorization": f"Bearer {nango_secret_key}"},
                params={"provider_config_key": cls.provider},
            )
            resp.raise_for_status()

            return resp.json().get("credentials", {})

    @classmethod
    async def is_connected(cls, data: dict, **kwargs) -> bool:
        """Get the status of the nango integration"""

        if "providers" in kwargs:
            return cls.provider in data and cls.provider in kwargs["providers"]

        return cls.provider in data

    @classmethod
    async def ensure_nango_integration(
        cls,
        ctx: RemoteExecutionContext,
        nango_secret_key: str,
    ) -> None:
        """Ensure the integration exists and is properly configured in Nango"""

        config = await cls.get_config(ctx)

        integration_data = {
            "unique_key": cls.provider,
            "provider": cls.provider,
            "display_name": cls.display_name,
            "credentials": config["credentials"],
        }

        headers = {
            "Authorization": f"Bearer {nango_secret_key}",
            "Content-Type": "application/json",
        }

        async with httpx.AsyncClient() as client:
            # First, try to get the existing integration
            try:
                existing_response = await client.get(
                    f"https://api.nango.dev/integrations/{cls.provider}",
                    headers=headers,
                    params={"include": "credentials"},
                )
                existing_response.raise_for_status()
                existing_integration = existing_response.json()["data"]

                # Check if configuration needs updating
                if any(
                    existing_integration["credentials"].get(k) != v for k, v in integration_data["credentials"].items()
                ):
                    update_response = await client.patch(
                        f"https://api.nango.dev/integrations/{cls.provider}",
                        json={"credentials": integration_data["credentials"]},
                        headers=headers,
                    )
                    update_response.raise_for_status()
            except httpx.HTTPStatusError as e:
                if e.response.status_code == 404:
                    logger.info(
                        f"Integration {cls.provider} not found in Nango, creating...",
                    )
                    create_response = await client.post(
                        "https://api.nango.dev/integrations",
                        json=integration_data,
                        headers=headers,
                    )
                    create_response.raise_for_status()
                else:
                    # Re-raise other HTTP errors
                    raise

    def connect(self, config: dict) -> None:
        """Connect to the integration"""
        self.config = config

    def disconnect(self) -> None:
        """Disconnect from the integration - use delete_integration for actual deletion"""



"""Node-related data models and schemas"""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel

from noxus_sdk.nodes.connector import AnyConnector


class SubflowConfig(BaseModel):
    workflow_id: str
    workflow_name: str
    workflow_nodes: list[str]


class NodeInput(BaseModel):
    name: str
    label: str
    definition: dict
    optional: bool


class NodeOutput(BaseModel):
    name: str
    label: str
    definition: dict
    optional: bool


class NodeDefinition(BaseModel):
    inputs: list[NodeInput]
    outputs: list[NodeOutput]
    config: dict
    type: str
    color: str
    image: str | None = None
    title: str | None = None
    description: str | None = None
    small_description: str | None = None
    documentation_url: str | None = None
    category: str | None = None
    sub_category: str | None = None
    example: str | None = None
    integrations: list[str] | None = None
    providers: list[str] | None = None
    knowledge_base_support: list[str] | None = None
    config_endpoint: str | None = None
    subflow_config: SubflowConfig | None = None
    is_valid: bool | None = None
    visible: bool = True
    is_available: bool = True
    show_to_user: bool = True
    paywalled: bool = False


class ConfigResponse(BaseModel):
    config: dict[str, dict]
    inputs: list[AnyConnector]
    outputs: list[AnyConnector]
    title: str | None = None
    ready: bool = False
    config_values: dict[str, Any] = {}

    def find_input(self, name: str) -> AnyConnector | None:
        for i in self.inputs:
            if i.name == name:
                return i
        return None


class ExecutionResponse(BaseModel):
    success: bool
    outputs: dict[str, Any]



from __future__ import annotations

from noxus_sdk.resources.base import BaseService
from noxus_sdk.workflows.agentflow import AgentFlowDefinition


class AgentFlowService(BaseService[AgentFlowDefinition]):
    async def alist(
        self,
        page: int = 1,
        page_size: int = 10,
    ) -> list[AgentFlowDefinition]:
        workflows_data = await self.client.apget(
            "/v1/workflows",
            params={"page": page, "page_size": page_size, "type": "agent_flow"},
            page=page,
            page_size=page_size,
        )
        return [AgentFlowDefinition.model_validate({"client": self.client, **data}) for data in workflows_data]

    def list(self, page: int = 1, page_size: int = 10) -> list[AgentFlowDefinition]:
        workflows_data = self.client.pget(
            "/v1/workflows",
            params={"page": page, "page_size": page_size, "type": "agent_flow"},
            page=page,
            page_size=page_size,
        )
        return [AgentFlowDefinition.model_validate({"client": self.client, **data}) for data in workflows_data]

    def delete(self, workflow_id: str) -> None:
        self.client.delete(f"/v1/workflows/{workflow_id}")

    async def adelete(self, workflow_id: str) -> None:
        await self.client.adelete(f"/v1/workflows/{workflow_id}")

    def save(self, workflow: AgentFlowDefinition) -> AgentFlowDefinition:
        w = self.client.post("/v1/workflows", workflow.to_noxus())
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    async def asave(self, workflow: AgentFlowDefinition) -> AgentFlowDefinition:
        w = await self.client.apost("/v1/workflows", workflow.to_noxus())
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    def get(self, workflow_id: str) -> AgentFlowDefinition:
        w = self.client.get(f"/v1/workflows/{workflow_id}")
        return AgentFlowDefinition.model_validate({"client": self.client, **w})

    async def aget(self, workflow_id: str) -> AgentFlowDefinition:
        w = await self.client.aget(f"/v1/workflows/{workflow_id}")
        return AgentFlowDefinition.model_validate({"client": self.client, **w})

    def update(
        self,
        workflow_id: str,
        workflow: AgentFlowDefinition,
        *,
        force: bool = False,
    ) -> AgentFlowDefinition:
        w = self.client.patch(
            f"/v1/workflows/{workflow_id}?force={force}",
            workflow.to_noxus(),
        )
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    async def aupdate(
        self,
        workflow_id: str,
        workflow: AgentFlowDefinition,
        *,
        force: bool = False,
    ) -> AgentFlowDefinition:
        w = await self.client.apatch(
            f"/v1/workflows/{workflow_id}?force={force}",
            workflow.to_noxus(),
        )
        workflow.refresh_from_data(client=self.client, **w)
        return workflow



from __future__ import annotations

import builtins
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, TypeAlias

import aiofiles
from pydantic import BaseModel, ConfigDict, Field

from noxus_sdk.resources.base import BaseResource, BaseService
from noxus_sdk.resources.runs import Run

if TYPE_CHECKING:
    from noxus_sdk.client import HttpxFile


RunStatus = Literal["queued", "running", "failed", "completed", "stopped"]
DocumentStatus = Literal["trained", "training", "error", "uploaded", "folder"]

SourceType = Literal[
    "document",
    "google_drive",
    "onedrive",
    "sharepoint",
    "website",
    "custom",
]

RunID: TypeAlias = str


class File(BaseModel):
    name: str
    size: int
    content_type: str
    source_type: str
    uri: str


class GoogleFile(BaseModel):
    id: str
    name: str
    mime_type: str
    size: int


class OneDriveFile(BaseModel):
    id: str
    name: str
    size: int
    web_url: str


class WebsiteWithDepth(BaseModel):
    url: str
    depth: int = 1


# Base document source config
class BaseDocumentSourceConfig(BaseModel):
    pass


# Regular document source config
class SpotFileConfig(BaseDocumentSourceConfig):
    files: list[File]


# Upload document source config
class UploadFileConfig(BaseDocumentSourceConfig):
    name: str
    content: bytes
    content_type: str


# Document source with discriminated union
class DocumentSourceConfig(BaseModel):
    files: builtins.list[File]


class DocumentSource(BaseModel):
    config: DocumentSourceConfig
    source_type: Literal["document"] = "document"
    subtype: str | None = None


class Source(BaseModel):
    source: DocumentSource


class KnowledgeBaseIngestion(BaseModel):
    batch_size: int
    default_chunk_size: int
    default_chunk_overlap: int
    enrich_chunks_mode: Literal["inject_summary", "contextual"] = "contextual"
    enrich_pre_made_qa: bool


class KnowledgeBaseRetrieval(BaseModel):
    type: Literal[
        "full_text_search",
        "semantic_search",
        "hybrid_search",
        "hybrid_reranking",
    ] = "hybrid_reranking"
    hybrid_settings: dict
    reranker_settings: dict


class KnowledgeBaseHybridSettings(BaseModel):
    fts_weight: float


class KnowledgeBaseSettings(BaseModel):
    ingestion: KnowledgeBaseIngestion
    retrieval: KnowledgeBaseRetrieval


class KBConfigV3(BaseModel):
    embedding_model: list[str] = Field(
        default=["vertexai/text-multilingual-embedding-002"],
        min_length=1,
    )
    default_chunk_size: int = 2048
    default_chunk_overlap: int = 512
    csv_row_as_document: bool = True


class KnowledgeBaseDocument(BaseModel):
    id: str
    name: str
    prefix: str
    status: DocumentStatus
    size: int
    source_type: str | None
    created_at: str
    updated_at: str
    error: dict | None = None


class CreateDocument(BaseModel):
    name: str
    prefix: str = "/"
    status: str = "uploaded"


class UpdateDocument(BaseModel):
    prefix: str | None = None
    status: DocumentStatus | None = None


class KnowledgeBase(BaseResource):
    model_config = ConfigDict(validate_assignment=True)

    id: str
    group_id: str
    name: str
    status: str
    description: str
    document_types: builtins.list[str]
    kb_type: str
    size: int
    num_docs: int
    created_at: str
    updated_at: str
    total_documents: int
    training_documents: int
    trained_documents: int
    error_documents: int
    uploaded_documents: int
    source_types: dict
    training_source_types: builtins.list[str]
    settings_: KnowledgeBaseSettings | KBConfigV3
    retrieval: dict | None = None
    error: dict | None = None
    embeddings: dict | None = None
    documents: builtins.list[KnowledgeBaseDocument] = []  # noqa: RUF012
    version: Literal["v2", "v3"] = "v3"

    def refresh(self) -> KnowledgeBase:
        response = self.client.get(f"/v1/knowledge-bases/{self.id}")
        for key, value in response.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    async def arefresh(self) -> KnowledgeBase:
        response = await self.client.aget(f"/v1/knowledge-bases/{self.id}")
        for key, value in response.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    def delete(self) -> bool:
        response = self.client.delete(f"/v1/knowledge-bases/{self.id}")
        return response["success"]

    async def adelete(self) -> bool:
        response = await self.client.adelete(f"/v1/knowledge-bases/{self.id}")
        return response["success"]

    def get_runs(
        self,
        status: RunStatus | None = None,
        run_ids: str | None = None,
    ) -> builtins.list[Run]:
        params: dict[str, str] = {}
        if status:
            params["status"] = status
        if run_ids:
            params["run_ids"] = run_ids

        response = self.client.get(f"/v1/knowledge-bases/{self.id}/runs", params=params)
        return [Run(client=self.client, **run) for run in response]

    async def aget_runs(
        self,
        status: RunStatus | None = None,
        run_ids: str | None = None,
    ) -> builtins.list[Run]:
        params: dict[str, str] = {}
        if status:
            params["status"] = status
        if run_ids:
            params["run_ids"] = run_ids

        response = await self.client.aget(
            f"/v1/knowledge-bases/{self.id}/runs",
            params=params,
        )
        return [Run(client=self.client, **run) for run in response]

    def get_document(self, document_id: str) -> KnowledgeBaseDocument:
        response = self.client.get(
            f"/v1/knowledge-bases/{self.id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    async def aget_document(self, document_id: str) -> KnowledgeBaseDocument:
        response = await self.client.aget(
            f"/v1/knowledge-bases/{self.id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    def create_document(self, document: CreateDocument) -> KnowledgeBaseDocument:
        response = self.client.post(
            f"/v1/knowledge-bases/{self.id}/document",
            body=document.model_dump(),
        )
        return KnowledgeBaseDocument(**response)

    async def acreate_document(self, document: CreateDocument) -> KnowledgeBaseDocument:
        response = await self.client.apost(
            f"/v1/knowledge-bases/{self.id}/document",
            body=document.model_dump(),
        )
        return KnowledgeBaseDocument(**response)

    def upload_document(
        self,
        files: builtins.list[str | Path],
        prefix: str = "/",
    ) -> builtins.list[RunID]:
        files_list: builtins.list[HttpxFile] = []
        for file in files:
            with open(str(file), "rb") as f:
                files_list.append(("files", (Path(file).name, f.read(), None)))

        return self.client.post(
            f"/v1/knowledge-bases/{self.id}/upload_train",
            files=files_list,
            params={"prefix": prefix},
        )

    async def aupload_document(
        self,
        files: builtins.list[str | Path],
        prefix: str = "/",
    ) -> builtins.list[RunID]:
        files_list: builtins.list[HttpxFile] = []
        for file in files:
            async with aiofiles.open(str(file), "rb") as f:
                content = await f.read()
                files_list.append(("files", (Path(file).name, content, None)))

        return await self.client.apost(
            f"/v1/knowledge-bases/{self.id}/upload_train",
            files=files_list,
            params={"prefix": prefix},
        )

    def update_document(
        self,
        document_id: str,
        update: UpdateDocument,
    ) -> KnowledgeBaseDocument:
        response = self.client.patch(
            f"/v1/knowledge-bases/{self.id}/document/{document_id}",
            update.model_dump(exclude_none=True),
        )
        return KnowledgeBaseDocument(**response)

    async def aupdate_document(
        self,
        document_id: str,
        update: UpdateDocument,
    ) -> KnowledgeBaseDocument:
        response = await self.client.apatch(
            f"/v1/knowledge-bases/{self.id}/document/{document_id}",
            update.model_dump(exclude_none=True),
        )
        return KnowledgeBaseDocument(**response)

    def delete_document(self, document_id: str) -> KnowledgeBaseDocument:
        response = self.client.delete(
            f"/v1/knowledge-bases/{self.id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    async def adelete_document(self, document_id: str) -> KnowledgeBaseDocument:
        response = await self.client.adelete(
            f"/v1/knowledge-bases/{self.id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    def list_documents(
        self,
        status: DocumentStatus,
        page: int = 1,
        page_size: int = 10,
    ) -> builtins.list[KnowledgeBaseDocument]:
        params: dict[str, Any] = {"page": page, "page_size": page_size}
        if status:
            params["status"] = status

        response = self.client.get(
            f"/v1/knowledge-bases/{self.id}/documents/{status}",
            params=params,
        )
        return [KnowledgeBaseDocument(**doc) for doc in response["items"]]

    async def alist_documents(
        self,
        status: DocumentStatus,
        page: int = 1,
        page_size: int = 10,
    ) -> builtins.list[KnowledgeBaseDocument]:
        params: dict[str, Any] = {"page": page, "page_size": page_size}
        if status:
            params["status"] = status

        response = await self.client.aget(
            f"/v1/knowledge-bases/{self.id}/documents/{status}",
            params=params,
        )
        return [KnowledgeBaseDocument(**doc) for doc in response["items"]]


class KnowledgeBaseService(BaseService[KnowledgeBase]):
    def list(self, page: int = 1, page_size: int = 10) -> builtins.list[KnowledgeBase]:
        knowledge_bases = self.client.pget(
            "/v1/knowledge-bases",
            params={"page": page, "page_size": page_size},
            page=page,
            page_size=page_size,
        )
        return [KnowledgeBase(client=self.client, **knowledge_base) for knowledge_base in knowledge_bases]

    async def alist(
        self,
        page: int = 1,
        page_size: int = 10,
    ) -> builtins.list[KnowledgeBase]:
        knowledge_bases = await self.client.apget(
            "/v1/knowledge-bases",
            params={"page": page, "page_size": page_size},
            page=page,
            page_size=page_size,
        )
        return [KnowledgeBase(client=self.client, **knowledge_base) for knowledge_base in knowledge_bases]

    def get(self, knowledge_base_id: str) -> KnowledgeBase:
        knowledge_base = self.client.get(f"/v1/knowledge-bases/{knowledge_base_id}")
        return KnowledgeBase(client=self.client, **knowledge_base)

    async def aget(self, knowledge_base_id: str) -> KnowledgeBase:
        knowledge_base = await self.client.aget(
            f"/v1/knowledge-bases/{knowledge_base_id}",
        )
        return KnowledgeBase(client=self.client, **knowledge_base)

    def create(
        self,
        name: str,
        description: str,
        document_types: builtins.list[str],
        settings_: KnowledgeBaseSettings | KBConfigV3,
        version: Literal["v2", "v3"] = "v3",
    ) -> KnowledgeBase:
        knowledge_base = self.client.post(
            "/v1/knowledge-bases",
            {
                "name": name,
                "description": description,
                "document_types": document_types,
                "settings_": settings_.model_dump(),
                "kb_type": "entity",
                "version": version,
            },
        )
        return KnowledgeBase(client=self.client, **knowledge_base)

    async def acreate(
        self,
        name: str,
        description: str,
        document_types: builtins.list[str],
        settings_: KnowledgeBaseSettings | KBConfigV3,
        version: Literal["v2", "v3"] = "v3",
    ) -> KnowledgeBase:
        knowledge_base = await self.client.apost(
            "/v1/knowledge-bases",
            {
                "name": name,
                "description": description,
                "document_types": document_types,
                "settings_": settings_.model_dump(),
                "kb_type": "entity",
                "version": version,
            },
        )

        return KnowledgeBase(client=self.client, **knowledge_base)

    def delete(self, knowledge_base_id: str) -> bool:
        response = self.client.delete(f"/v1/knowledge-bases/{knowledge_base_id}")
        return response["success"]

    async def adelete(self, knowledge_base_id: str) -> bool:
        response = await self.client.adelete(f"/v1/knowledge-bases/{knowledge_base_id}")
        return response["success"]

    def get_runs(
        self,
        knowledge_base_id: str,
        status: RunStatus | None = None,
        run_ids: str | None = None,
    ) -> builtins.list[Run]:
        params: dict[str, str] = {}
        if status:
            params["status"] = status
        if run_ids:
            params["run_ids"] = run_ids

        response = self.client.get(
            f"/v1/knowledge-bases/{knowledge_base_id}/runs",
            params=params,
        )
        return [Run(client=self.client, **run) for run in response]

    async def aget_runs(
        self,
        knowledge_base_id: str,
        status: RunStatus | None = None,
        run_ids: str | None = None,
    ) -> builtins.list[Run]:
        params: dict[str, str] = {}
        if status:
            params["status"] = status
        if run_ids:
            params["run_ids"] = run_ids

        response = await self.client.aget(
            f"/v1/knowledge-bases/{knowledge_base_id}/runs",
            params=params,
        )
        return [Run(client=self.client, **run) for run in response]

    def get_document(
        self,
        knowledge_base_id: str,
        document_id: str,
    ) -> KnowledgeBaseDocument:
        response = self.client.get(
            f"/v1/knowledge-bases/{knowledge_base_id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    async def aget_document(
        self,
        knowledge_base_id: str,
        document_id: str,
    ) -> KnowledgeBaseDocument:
        response = await self.client.aget(
            f"/v1/knowledge-bases/{knowledge_base_id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    def update_document(
        self,
        knowledge_base_id: str,
        document_id: str,
        update: UpdateDocument,
    ) -> KnowledgeBaseDocument:
        response = self.client.patch(
            f"/v1/knowledge-bases/{knowledge_base_id}/document/{document_id}",
            update.model_dump(exclude_none=True),
        )
        return KnowledgeBaseDocument(**response)

    async def aupdate_document(
        self,
        knowledge_base_id: str,
        document_id: str,
        update: UpdateDocument,
    ) -> KnowledgeBaseDocument:
        response = await self.client.apatch(
            f"/v1/knowledge-bases/{knowledge_base_id}/document/{document_id}",
            update.model_dump(exclude_none=True),
        )
        return KnowledgeBaseDocument(**response)

    def delete_document(
        self,
        knowledge_base_id: str,
        document_id: str,
    ) -> KnowledgeBaseDocument:
        response = self.client.delete(
            f"/v1/knowledge-bases/{knowledge_base_id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    async def adelete_document(
        self,
        knowledge_base_id: str,
        document_id: str,
    ) -> KnowledgeBaseDocument:
        response = await self.client.adelete(
            f"/v1/knowledge-bases/{knowledge_base_id}/document/{document_id}",
        )
        return KnowledgeBaseDocument(**response)

    def list_documents(
        self,
        knowledge_base_id: str,
        status: DocumentStatus,
        page: int = 1,
        page_size: int = 10,
    ) -> builtins.list[KnowledgeBaseDocument]:
        params: dict[str, Any] = {"page": page, "page_size": page_size}
        if status:
            params["status"] = status

        response = self.client.get(
            f"/v1/knowledge-bases/{knowledge_base_id}/documents/{status}",
            params=params,
        )
        return [KnowledgeBaseDocument(**doc) for doc in response]

    async def alist_documents(
        self,
        knowledge_base_id: str,
        status: DocumentStatus,
        page: int = 1,
        page_size: int = 10,
    ) -> builtins.list[KnowledgeBaseDocument]:
        params: dict[str, Any] = {"page": page, "page_size": page_size}
        if status:
            params["status"] = status

        response = await self.client.aget(
            f"/v1/knowledge-bases/{knowledge_base_id}/documents/{status}",
            params=params,
        )
        return [KnowledgeBaseDocument(**doc) for doc in response]

    def create_document(
        self,
        knowledge_base_id: str,
        document: CreateDocument,
    ) -> KnowledgeBaseDocument:
        response = self.client.post(
            f"/v1/knowledge-bases/{knowledge_base_id}/document",
            body=document.model_dump(),
        )
        return KnowledgeBaseDocument(**response)

    async def acreate_document(
        self,
        knowledge_base_id: str,
        document: CreateDocument,
    ) -> KnowledgeBaseDocument:
        response = await self.client.apost(
            f"/v1/knowledge-bases/{knowledge_base_id}/document",
            body=document.model_dump(),
        )
        return KnowledgeBaseDocument(**response)

    def train_document(
        self,
        knowledge_base_id: str,
        source: Source,
        prefix: str = "/",
    ) -> builtins.list[RunID]:
        return self.client.post(
            f"/v1/knowledge-bases/{knowledge_base_id}/generic_train",
            body=source.model_dump(),
            params={"prefix": prefix},
        )

    async def atrain_document(
        self,
        knowledge_base_id: str,
        source: Source,
        prefix: str = "/",
    ) -> builtins.list[RunID]:
        return await self.client.apost(
            f"/v1/knowledge-bases/{knowledge_base_id}/generic_train",
            body=source.model_dump(),
            params={"prefix": prefix},
        )

    def upload_document(
        self,
        knowledge_base_id: str,
        files: builtins.list[str | Path],
        prefix: str = "/",
    ) -> builtins.list[RunID]:
        files_list: builtins.list[HttpxFile] = []
        for file in files:
            with open(str(file), "rb") as f:
                files_list.append(("files", (Path(file).name, f.read(), None)))

        return self.client.post(
            f"/v1/knowledge-bases/{knowledge_base_id}/upload_train",
            files=files_list,
            params={"prefix": prefix},
        )

    async def aupload_document(
        self,
        knowledge_base_id: str,
        files: builtins.list[str | Path],
        prefix: str = "/",
    ) -> builtins.list[RunID]:
        files_list: builtins.list[HttpxFile] = []
        for file in files:
            async with aiofiles.open(str(file), "rb") as f:
                content = await f.read()
                files_list.append(("files", (Path(file).name, content, None)))

        return await self.client.apost(
            f"/v1/knowledge-bases/{knowledge_base_id}/upload_train",
            files=files_list,
            params={"prefix": prefix},
        )



import subprocess
from pathlib import Path

from cookiecutter.main import cookiecutter


def create_plugin(output_dir: Path = Path()) -> Path:
    """Create a new plugin using cookiecutter template"""

    template_path = Path(__file__).parent.parent / "templates" / "cookiecutter-plugin"

    try:
        path = cookiecutter(
            str(template_path),
            no_input=False,
            output_dir=str(output_dir),
        )

        if not path:
            raise RuntimeError("No plugin directory was created")

        return Path(path)
    except subprocess.CalledProcessError as e:
        if e.returncode == 1:
            # User cancelled, not an error
            raise KeyboardInterrupt("Plugin creation cancelled by user") from None
        raise RuntimeError(f"Cookiecutter failed: {e}") from e



from __future__ import annotations

import enum
from typing import TYPE_CHECKING, TypeAlias

from pydantic import BaseModel, ConfigDict, Field

from noxus_sdk.resources.base import BaseResource, BaseService
from noxus_sdk.resources.conversations import (
    ConversationSettings,
)

if TYPE_CHECKING:
    from uuid import UUID

AgentSettings: TypeAlias = ConversationSettings


class TriggerType(str, enum.Enum):
    SLACK = "slack"
    TEAMS = "teams"


class TriggerData(BaseModel):
    trigger_type: TriggerType = Field(exclude=True)
    team_id: str
    channel: str | None = None
    keyword: str | None = None


class AssistantTrigger(BaseResource):
    id: UUID
    group_id: UUID
    definition: dict
    routing_key: str
    agent_id: UUID = Field(alias="assistant_id")

    model_config = ConfigDict(from_attributes=True)

    def delete(self) -> None:
        self.client.delete(f"/v1/triggers/{self.id}")

    async def adelete(self) -> None:
        await self.client.adelete(f"/v1/triggers/{self.id}")


class Agent(BaseResource):
    id: str
    name: str
    definition: AgentSettings
    draft_definition: AgentSettings | None = None
    model_config = ConfigDict(validate_assignment=True, extra="allow")

    def add_trigger(self, trigger_data: TriggerData) -> AssistantTrigger:
        url = f"/v1/agents/{self.id}/triggers/{trigger_data.trigger_type.value}"
        result = self.client.post(url, trigger_data.model_dump())
        return AssistantTrigger(client=self.client, **result)

    async def aadd_trigger(self, trigger_data: TriggerData) -> AssistantTrigger:
        url = f"/v1/agents/{self.id}/triggers/{trigger_data.trigger_type.value}"
        result = await self.client.apost(url, trigger_data.model_dump())
        return AssistantTrigger(client=self.client, **result)

    def triggers(self) -> list[AssistantTrigger]:
        result = self.client.get(f"/v1/agents/{self.id}/triggers")
        return [AssistantTrigger(client=self.client, **result) for result in result]

    async def atriggers(self) -> list[AssistantTrigger]:
        result = await self.client.aget(f"/v1/agents/{self.id}/triggers")
        return [AssistantTrigger(client=self.client, **result) for result in result]

    def update(
        self,
        name: str,
        settings: AgentSettings,
        *,
        preview: bool = False,
    ) -> Agent:
        result = self.client.patch(
            f"/v1/agents/{self.id}",
            {"name": name, "definition": settings.model_dump()},
            params={"preview": preview},
        )
        for key, value in result.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    def delete(self) -> None:
        self.client.delete(f"/v1/agents/{self.id}")


class AgentService(BaseService[Agent]):
    async def alist(self) -> list[Agent]:
        results = await self.client.apget("/v1/agents")
        return [Agent(client=self.client, **result) for result in results]

    def list(self) -> list[Agent]:
        results = self.client.pget("/v1/agents")
        return [Agent(client=self.client, **result) for result in results]

    def create(self, name: str, settings: AgentSettings) -> Agent:
        result = self.client.post(
            "/v1/agents",
            {"name": name, "definition": settings.model_dump()},
        )
        return Agent(client=self.client, **result)

    async def acreate(self, name: str, settings: AgentSettings) -> Agent:
        result = await self.client.apost(
            "/v1/agents",
            {"name": name, "definition": settings.model_dump()},
        )
        return Agent(client=self.client, **result)

    def get(self, agent_id: str) -> Agent:
        result = self.client.get(f"/v1/agents/{agent_id}")
        return Agent(client=self.client, **result)

    async def aget(self, agent_id: str) -> Agent:
        result = await self.client.aget(f"/v1/agents/{agent_id}")
        return Agent(client=self.client, **result)

    def update(
        self,
        agent_id: str,
        name: str | None = None,
        settings: AgentSettings | None = None,
        *,
        preview: bool = False,
    ) -> Agent:
        result = self.client.patch(
            f"/v1/agents/{agent_id}",
            {"name": name, "definition": settings.model_dump() if settings else None},
            params={"preview": preview},
        )
        return Agent(client=self.client, **result)

    async def aupdate(
        self,
        agent_id: str,
        name: str | None = None,
        settings: AgentSettings | None = None,
        *,
        preview: bool = False,
    ) -> Agent:
        result = await self.client.apatch(
            f"/v1/agents/{agent_id}",
            {
                "name": name,
                "definition": settings.model_dump() if settings else None,
            },
            params={"preview": preview},
        )
        return Agent(client=self.client, **result)

    def delete(self, agent_id: str) -> None:
        self.client.delete(f"/v1/agents/{agent_id}")

    async def adelete(self, agent_id: str) -> None:
        await self.client.adelete(f"/v1/agents/{agent_id}")



"""Remote execution context for plugins"""

from __future__ import annotations

from pydantic import BaseModel


class RemoteExecutionContext(BaseModel):
    plugin_config: dict = {}
    integration_credentials: dict[str, dict] = {}

    def get_integration_credentials(self, integration_name: str) -> dict:
        return self.integration_credentials[integration_name] if self.integration_credentials else {}



import os
import uuid
from pathlib import Path

import pytest
from filelock import FileLock

from noxus_sdk.client import Client
from noxus_sdk.resources.knowledge_bases import (
    KBConfigV3,
)


@pytest.fixture
def anyio_backend():
    return "asyncio"


@pytest.fixture(scope="session")
def workspace_client():
    client = Client(
        os.environ.get("NOXUS_API_KEY", ""),
        base_url=os.environ.get("NOXUS_BASE_URL", "https://backend.noxus.ai"),
    )

    fn = Path(".workspace_lock")
    with FileLock(str(fn) + ".lock") as lock:
        if fn.is_file():
            pass
        else:
            for workspace in client.admin.list_workspaces():
                if workspace.name.startswith("sdk-"):
                    print("Deleting", workspace.name)
                    workspace.delete()
            fn.touch()

    return client


@pytest.fixture
def api_key(workspace_client: Client):
    workspace = workspace_client.admin.create_workspace(f"sdk-{uuid.uuid4()}")
    api_key = workspace.add_api_key("test_key")
    yield api_key.value
    workspace.delete()


@pytest.fixture
def client(api_key: str):
    return Client(
        api_key,
        base_url=os.environ.get("NOXUS_BASE_URL", "https://backend.noxus.ai"),
    )


@pytest.fixture
async def test_file():
    f = Path("test_file.txt")
    f.write_text("Test content for document upload")

    yield f

    try:
        f.unlink()
    except Exception:
        pass


@pytest.fixture
async def kb(client: Client, test_file: Path):
    settings = KBConfigV3()

    kb = await client.knowledge_bases.acreate(
        name="test_kb",
        description="Test Knowledge Base",
        document_types=["text"],
        settings_=settings,
    )
    # await kb.aupload_document([test_file], prefix="/test1")

    yield kb

    try:
        await kb.adelete()
    except Exception:
        pass



from __future__ import annotations

import sys
from contextlib import contextmanager
from importlib import import_module
from inspect import getmembers, isclass
from typing import TYPE_CHECKING

from loguru import logger

from noxus_sdk.plugins import BasePlugin, PluginManifest
from noxus_sdk.schemas import ValidationResult

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from types import ModuleType


@contextmanager
def _prepend_sys_path(path: Path) -> Iterable[None]:
    """Temporarily prepend a path to sys.path."""
    inserted = False
    path_str = str(path)
    if path_str not in sys.path:
        sys.path.insert(0, path_str)
        inserted = True
    try:
        yield
    finally:
        if inserted and path_str in sys.path:
            sys.path.remove(path_str)


def _iter_plugin_classes(module: ModuleType) -> list[type[BasePlugin]]:
    """Return BasePlugin subclasses defined in the given module."""
    plugin_classes: list[type[BasePlugin]] = []
    for _, obj in getmembers(module, isclass):
        if issubclass(obj, BasePlugin) and obj is not BasePlugin and obj.__module__ == module.__name__:
            plugin_classes.append(obj)
    return plugin_classes


def _is_valid_python_module_name(name: str) -> bool:
    """Check if a name is a valid Python module name"""
    if not name or not name.isidentifier():
        return False
    # Additional check for reserved keywords
    import keyword

    return not keyword.iskeyword(name)


def _find_plugin_package(path: Path) -> Path | None:
    """Find the actual plugin package directory (with __init__.py)"""
    # Convert to absolute path to ensure consistent behavior
    path = path.resolve()
    logger.debug(f"Finding plugin package in {path}")

    # First, scan for any subdirectory that contains __init__.py with a valid module name
    try:
        for item in path.iterdir():
            if (
                item.is_dir()
                and (item / "__init__.py").exists()
                and _is_valid_python_module_name(item.name)
                and item.name != "tests"
            ):
                # Found a valid Python package directory
                return item
    except (OSError, PermissionError):
        # Handle cases where we can't read the directory
        pass

    # Then, check the traditional structure where package has same name as parent
    nested_dir = path / path.name
    if (nested_dir / "__init__.py").exists() and _is_valid_python_module_name(
        path.name,
    ):
        return nested_dir

    # Finally, check if the path itself is a Python package with a valid name
    if (path / "__init__.py").exists() and _is_valid_python_module_name(path.name):
        return path

    return None


def _import_and_find_plugin(
    package_dir: Path,
    root_dir: Path,
) -> type[BasePlugin] | None:
    """Import package and find the BasePlugin subclass."""
    package_name = package_dir.name

    with _prepend_sys_path(root_dir):
        # Import the package using standard import semantics
        module = import_module(package_name)

        plugin_classes = _iter_plugin_classes(module)

        if not plugin_classes:
            return None
        if len(plugin_classes) == 1:
            return plugin_classes[0]
        if len(plugin_classes) > 1:
            class_names = ", ".join(cls.__name__ for cls in plugin_classes)
            raise ValueError(
                f"Multiple plugin classes found in __init__.py: {class_names}. Use only one plugin class.",
            )

        return None


def discover_and_load_plugin(
    path: Path,
) -> tuple[type[BasePlugin] | None, ValidationResult]:
    """Discover plugin class in a plugin package"""
    errors = []
    warnings = []

    plugin_package_dir = _find_plugin_package(path)
    logger.debug(f"Found plugin package in {plugin_package_dir}")

    if not plugin_package_dir:
        errors.append(
            f"No valid Python package found. Expected __init__.py in {path}, {path / path.name}, or any subdirectory",
        )
        return None, ValidationResult(valid=False, errors=errors, warnings=warnings)

    # Import the package and find BasePlugin subclasses
    plugin_class = _import_and_find_plugin(plugin_package_dir, path)
    if not plugin_class:
        errors.append(
            "No BasePlugin subclasses found in package. Define your plugin class that inherits from BasePlugin.",
        )
        return None, ValidationResult(valid=False, errors=errors, warnings=warnings)

    return plugin_class, ValidationResult(valid=True, errors=errors, warnings=warnings)


def validate_plugin(path: Path) -> tuple[PluginManifest | None, ValidationResult]:
    """Validate everything in the plugin to ensure it properly implements the plugin interface"""

    # First discover and load the plugin class
    plugin_class, discovery_result = discover_and_load_plugin(path)

    if discovery_result.errors or plugin_class is None:
        return None, discovery_result

    # Generate manifest from class
    manifest = None
    validation_errors = []

    try:
        manifest = plugin_class.get_manifest()
    except Exception as e:  # noqa: BLE001 - If the plugin manifest generation code fails, we want to return a validation result with the error. We dont control the code so need to catch all exceptions.
        validation_errors.append(f"Failed to generate manifest: {e}")

    # Combine results
    all_errors = (discovery_result.errors or []) + validation_errors
    all_warnings = discovery_result.warnings or []

    return manifest, ValidationResult(
        valid=True,
        errors=all_errors,
        warnings=all_warnings,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from noxus_sdk.resources.base import BaseResource, BaseService
from noxus_sdk.workflows import WorkflowDefinition

if TYPE_CHECKING:
    import builtins
    from datetime import datetime
    from uuid import UUID


class WorkflowVersion(BaseResource):
    id: UUID
    name: str
    description: str | None = None
    created_at: datetime
    created_by: UUID | None = None
    definition: dict


class WorkflowService(BaseService[WorkflowDefinition]):
    async def alist(
        self,
        page: int = 1,
        page_size: int = 10,
    ) -> list[WorkflowDefinition]:
        workflows_data = await self.client.apget(
            "/v1/workflows",
            params={"page": page, "page_size": page_size, "type": "flow"},
            page=page,
            page_size=page_size,
        )
        return [WorkflowDefinition.model_validate({"client": self.client, **data}) for data in workflows_data]

    def list(self, page: int = 1, page_size: int = 10) -> list[WorkflowDefinition]:
        workflows_data = self.client.pget(
            "/v1/workflows",
            params={"page": page, "page_size": page_size, "type": "flow"},
            page=page,
            page_size=page_size,
        )
        return [WorkflowDefinition.model_validate({"client": self.client, **data}) for data in workflows_data]

    def delete(self, workflow_id: str) -> None:
        self.client.delete(f"/v1/workflows/{workflow_id}")

    async def adelete(self, workflow_id: str) -> None:
        await self.client.adelete(f"/v1/workflows/{workflow_id}")

    def save(self, workflow: WorkflowDefinition) -> WorkflowDefinition:
        w = self.client.post("/v1/workflows", workflow.to_noxus())
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    async def asave(self, workflow: WorkflowDefinition) -> WorkflowDefinition:
        w = await self.client.apost("/v1/workflows", workflow.to_noxus())
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    def get(self, workflow_id: str) -> WorkflowDefinition:
        w = self.client.get(f"/v1/workflows/{workflow_id}")
        return WorkflowDefinition.model_validate({"client": self.client, **w})

    async def aget(self, workflow_id: str) -> WorkflowDefinition:
        w = await self.client.aget(f"/v1/workflows/{workflow_id}")
        return WorkflowDefinition.model_validate({"client": self.client, **w})

    def update(
        self,
        workflow_id: str,
        workflow: WorkflowDefinition,
        *,
        force: bool = False,
    ) -> WorkflowDefinition:
        w = self.client.patch(
            f"/v1/workflows/{workflow_id}?force={force}",
            workflow.to_noxus(),
        )
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    async def aupdate(
        self,
        workflow_id: str,
        workflow: WorkflowDefinition,
        *,
        force: bool = False,
    ) -> WorkflowDefinition:
        w = await self.client.apatch(
            f"/v1/workflows/{workflow_id}?force={force}",
            workflow.to_noxus(),
        )
        workflow.refresh_from_data(client=self.client, **w)
        return workflow

    def save_version(
        self,
        workflow_id: str,
        workflow: WorkflowDefinition,
        name: str,
        description: str | None,
    ) -> WorkflowVersion:
        body = {
            "name": name,
            "description": description,
            "definition": workflow.to_noxus()["definition"],
        }
        w = self.client.post(
            f"/v1/workflows/{workflow_id}/versions",
            body,
        )
        return WorkflowVersion.model_validate({"client": self.client, **w})

    async def asave_version(
        self,
        workflow_id: str,
        workflow: WorkflowDefinition,
        name: str,
        description: str | None,
    ) -> WorkflowVersion:
        body = {
            "name": name,
            "description": description,
            "definition": workflow.to_noxus()["definition"],
        }
        w = await self.client.apost(
            f"/v1/workflows/{workflow_id}/versions",
            body,
        )
        return WorkflowVersion.model_validate({"client": self.client, **w})

    def list_versions(self, workflow_id: str) -> builtins.list[WorkflowVersion]:
        w = self.client.get(f"/v1/workflows/{workflow_id}/versions")
        return [WorkflowVersion.model_validate({"client": self.client, **v}) for v in w]

    async def alist_versions(self, workflow_id: str) -> builtins.list[WorkflowVersion]:
        w = await self.client.aget(f"/v1/workflows/{workflow_id}/versions")
        return [WorkflowVersion.model_validate({"client": self.client, **v}) for v in w]

    def update_version(
        self,
        workflow_id: str,
        version_id: str,
        name: str,
        description: str | None,
        definition: WorkflowDefinition,
    ) -> WorkflowVersion:
        w = self.client.patch(
            f"/v1/workflows/{workflow_id}/versions/{version_id}",
            {
                "name": name,
                "description": description,
                "definition": definition.to_noxus()["definition"],
            },
        )
        return WorkflowVersion.model_validate({"client": self.client, **w})

    async def aupdate_version(
        self,
        workflow_id: str,
        version_id: str,
        name: str,
        description: str | None,
        definition: WorkflowDefinition,
    ) -> WorkflowVersion:
        w = await self.client.apatch(
            f"/v1/workflows/{workflow_id}/versions/{version_id}",
            {
                "name": name,
                "description": description,
                "definition": definition.to_noxus()["definition"],
            },
        )
        return WorkflowVersion.model_validate({"client": self.client, **w})



"""{{ cookiecutter.description }}"""

from typing import Type

from noxus_sdk.plugins import BasePlugin, PluginConfiguration
from noxus_sdk.nodes import BaseNode

from {{ cookiecutter.__package_name }}.nodes import ExampleNode


class {{ cookiecutter.__package_name.replace('-', ' ').title().replace(' ', '') }}Configuration(PluginConfiguration):
    """Configuration for {{ cookiecutter.__package_name }}"""
    pass


class {{ cookiecutter.__package_name.replace('-', ' ').title().replace(' ', '') }}Plugin(BasePlugin[{{ cookiecutter.__package_name.replace('-', ' ').title().replace(' ', '') }}Configuration]):
    """{{ cookiecutter.description }}"""

    # Plugin metadata (auto-detected from package if not set)
    name = "{{ cookiecutter.__package_name }}"
    display_name = "{{ cookiecutter.plugin_name }}"
    version = "0.1.0"
    description = "{{ cookiecutter.description }}"
    author = "{{ cookiecutter.author_name }}"

    def nodes(self) -> list[Type[BaseNode]]:
        """Return the nodes provided by this plugin"""
        return [ExampleNode]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from pydantic import BaseModel

from noxus_sdk.integrations.schemas import IntegrationDefinition
from noxus_sdk.nodes.schemas import NodeDefinition
from noxus_sdk.plugins.types import PluginCategory

if TYPE_CHECKING:
    from pathlib import Path


class PluginManifest(BaseModel):
    """Complete plugin specification combining manifest and spec"""

    # Core plugin metadata
    name: str
    display_name: str
    version: str
    description: str
    category: PluginCategory = PluginCategory.OTHER
    author: str

    # Configuration
    config: dict

    # Execution configuration
    execution: Literal["runtime", "docker", "remote"] = "runtime"
    image: str | None = None
    endpoint: str | None = None

    # Plugin components
    nodes: list[NodeDefinition] = []
    integrations: list[IntegrationDefinition] = []

    @classmethod
    def from_file(cls, file_path: Path) -> PluginManifest:
        """Load plugin manifest from a file"""
        with open(file_path) as f:
            return cls.model_validate_json(f.read())



"""Shared utilities across noxus-sdk"""

from noxus_sdk.utils.errors import (
    NodeDefinitionError,
    NoxusError,
    PluginValidationError,
    UnexpectedError,
)
from noxus_sdk.utils.logging import setup_logging

__all__ = [
    "NodeDefinitionError",
    "NoxusError",
    "PluginValidationError",
    "UnexpectedError",
    "setup_logging",
]



"""Plugin-specific types and enums"""

from enum import Enum


class PluginExecutionMode(str, Enum):
    RUNTIME = "runtime"
    DOCKER = "docker"
    REMOTE = "remote"


class PluginCategory(str, Enum):
    GENERAL = "general"
    DOCUMENT = "document"
    OTHER = "other"


class PluginStatus(str, Enum):
    INSTALLING = "installing"
    MISSING_CONFIG = "missing_config"
    RUNNING = "running"
    RESTARTING = "restarting"
    UNINSTALLING = "uninstalling"
    UNINSTALLED = "uninstalled"
    ERROR = "error"


class PluginSourceType(str, Enum):
    GIT = "git"
    LOCAL = "local"
    UPLOAD = "upload"
    MARKETPLACE = "marketplace"



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal

from noxus_sdk.integrations.schemas import IntegrationDefinition

if TYPE_CHECKING:
    from uuid import UUID

    from noxus_sdk.plugins.context import RemoteExecutionContext


class BaseIntegration(ABC):
    """Base class for all integrations"""

    type: Literal["nango"]
    name: str
    display_name: str
    image: str
    description: str | None = None
    scopes: list[str]  # Will be set to an empty list if not set
    properties: dict  # Will be set to an empty dict if not set

    @classmethod
    def __init__subclass__(cls) -> None:
        if not cls.scopes:
            cls.scopes = []

        if not cls.properties:
            cls.properties = {}

    @classmethod
    def get_trigger_endpoint(cls, group_id: UUID) -> str | None:  # noqa: ARG003 - Here for documentation purposes
        """Get the trigger endpoint of the integration"""
        return None

    @classmethod
    def get_delete_endpoint(cls, group_id: UUID) -> str | None:  # noqa: ARG003 - Here for documentation purposes
        """Get the delete endpoint of the integration"""
        return None

    @classmethod
    def get_confirm_endpoint(cls, group_id: UUID) -> str | None:  # noqa: ARG003 - Here for documentation purposes
        """Get the confirm endpoint of the integration"""
        return None

    @classmethod
    async def get_config(cls, ctx: RemoteExecutionContext) -> dict:  # noqa: ARG003 - Here for documentation purposes
        """Get the config of the integration"""
        return {}

    @classmethod
    @abstractmethod
    async def is_connected(cls, data: dict, **kwargs) -> bool:
        """Get the status of the integration"""
        raise NotImplementedError("Integration must implement is_connected() method")

    @classmethod
    @abstractmethod
    async def get_credentials(cls, data: dict) -> dict:
        """Get the status of the integration"""
        return data

    @abstractmethod
    def connect(self, config: dict) -> None:
        """Connect to the integration"""
        raise NotImplementedError("Integration must implement connect() method")

    @abstractmethod
    def disconnect(self) -> None:
        """Get the auth type of the integration"""
        raise NotImplementedError("Integration must implement disconnect() method")

    @abstractmethod
    def update_config(self, config: dict) -> None:
        """Update the config of the integration"""
        raise NotImplementedError("Integration must implement update_config() method")

    @classmethod
    def get_definition(cls) -> IntegrationDefinition:
        """Convert integration class to IntegrationDefinition"""

        return IntegrationDefinition(
            name=cls.name,
            type=cls.type,  # type: ignore
            display_name=cls.display_name,
            description=cls.description,
            image=cls.image,
            scopes=cls.scopes,
            properties=cls.properties,
        )



"""Local plugin source for copying plugins from local directories"""

from __future__ import annotations

import json
import shutil
import tempfile
from pathlib import Path
from typing import Literal

import aiofiles
from loguru import logger
from pydantic import BaseModel, Field

from noxus_sdk.plugins.interfaces import PluginSource
from noxus_sdk.plugins.manifest import PluginManifest


class LocalPluginSource(PluginSource, BaseModel):
    """Local plugin source"""

    type: Literal["local"] = "local"
    path: str = Field(..., description="Local path to the plugin directory")

    def get_name(self) -> str:
        """Get the name of the plugin"""
        return Path(self.path).name

    async def download_plugin(self, output_dir: str | Path | None = None) -> Path:
        """Copy the plugin from local directory"""
        if output_dir is None:
            raise ValueError("Output directory must be specified")

        source_path = Path(self.path)
        if not source_path.exists():
            raise FileNotFoundError(f"Plugin directory {self.path} does not exist")

        if not source_path.is_dir():
            raise ValueError(f"Plugin path {self.path} is not a directory")

        target_name = source_path.name
        target_path = Path(output_dir) / target_name

        if target_path.exists():
            raise FileExistsError(f"Destination {target_path} already exists")

        logger.info(f"Copying plugin from {source_path} to {target_path}")

        shutil.copytree(source_path, target_path)

        return target_path

    async def get_manifest(self) -> PluginManifest:
        """Get the manifest of the plugin by copying to temp dir"""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Use download_plugin to get the plugin files
            plugin_path = await self.download_plugin(temp_dir)

            # Read manifest.json from copied plugin
            manifest_file = plugin_path / "manifest.json"
            if not manifest_file.exists():
                raise FileNotFoundError(f"manifest.json not found in {plugin_path}")

            # Load and parse manifest
            async with aiofiles.open(manifest_file) as f:
                content = await f.read()
                manifest_data = json.loads(content)

            return PluginManifest(**manifest_data)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Literal
from uuid import UUID, uuid4

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Discriminator,
    Field,
    ValidationError,
    model_validator,
)

from noxus_sdk.resources.base import BaseResource, BaseService

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator


class ConversationTool(BaseModel):
    type: str
    enabled: bool = True
    extra_instructions: str | None = None


class WebResearchTool(ConversationTool):
    """Tool that allows the user to search the web for information"""

    type: Literal["web_research"] = "web_research"


class NoxusQaTool(ConversationTool):
    """Tool that allows the user to answer questions about the Noxus platform"""

    type: Literal["noxus_qa"] = "noxus_qa"


class AttachFileTool(ConversationTool):
    """Tool that allows attaching files to a conversation"""

    type: Literal["attach_file"] = "attach_file"


class KnowledgeBaseSelectorTool(ConversationTool):
    """Tool that allows the user to select a knowledge base to answer questions about"""

    type: Literal["kb_selector"] = "kb_selector"


class KnowledgeBaseQaTool(ConversationTool):
    """Tool that allows the user to answer questions about a specific pre-selected knowledge base"""

    type: Literal["kb_qa"] = "kb_qa"
    kb_id: str


class HumanInTheLoopTool(ConversationTool):
    """Tool that allows the agent to escalate to human"""

    type: Literal["human_in_the_loop"] = "human_in_the_loop"


class WorkflowTool(ConversationTool):
    """Tool that allows the user to run a workflow"""

    type: Literal["workflow"] = "workflow"
    workflow_id: str


AnyToolSettings = Annotated[
    WebResearchTool
    | NoxusQaTool
    | KnowledgeBaseSelectorTool
    | KnowledgeBaseQaTool
    | WorkflowTool
    | HumanInTheLoopTool
    | AttachFileTool,
    Discriminator("type"),
]


class ConversationSettings(BaseModel):
    model: list[str] = Field(validation_alias=AliasChoices("model", "model_selection"))
    temperature: float
    max_tokens: int = 64000
    tools: list[AnyToolSettings]
    extra_instructions: str | None = None
    agent_flow_id: str | None = None


class ConversationFile(BaseModel):
    status: Literal["success"] = "success"
    name: str
    b64_content: str | None = None
    url: str | None = None
    id: str = Field(default_factory=lambda: str(uuid4()))
    size: int = 1
    type: str = ""

    @model_validator(mode="after")
    def validate_content_url(self) -> ConversationFile:
        if self.b64_content is None and self.url is None:
            raise ValidationError("Either base64 content or url must be provided")
        return self


class MessageRequest(BaseModel):
    content: str
    tool: Literal["web_research", "kb_qa", "workflow"] | str | None = None  # noqa: PYI051 - For documentation purposes
    kb_id: str | None = None
    workflow_id: str | None = None
    files: list[ConversationFile] | None = None
    model_selection: list[str] | None = None


class Message(BaseModel):
    id: UUID
    created_at: datetime
    message_parts: list[dict]


class Conversation(BaseResource):
    model_config = ConfigDict(validate_assignment=True)

    id: str
    name: str
    created_at: str
    last_updated_at: str
    settings: ConversationSettings
    etag: str | None = None
    messages: list[Message] = []  # noqa: RUF012
    status: str
    agent_id: str | None = Field(
        default=None,
        validation_alias=AliasChoices("assistant_id", "agent_id"),
    )

    def _update_w_response(self, response: dict) -> None:
        for key, value in response.items():
            if key == "assistant_id":
                key = "agent_id"  # noqa: PLW2901
            if hasattr(self, key):
                setattr(self, key, value)

    def refresh(self) -> Conversation:
        response = self.client.get(f"/v1/conversations/{self.id}")
        self._update_w_response(response)
        return self

    async def arefresh(self) -> Conversation:
        response = await self.client.aget(f"/v1/conversations/{self.id}")
        self._update_w_response(response)
        return self

    async def aget_messages(self) -> list[Message]:
        response = await self.arefresh()
        return [Message.model_validate(msg) for msg in response.messages]

    def get_messages(self) -> list[Message]:
        response = self.refresh()
        return [Message.model_validate(msg) for msg in response.messages]

    async def aadd_message(self, message: MessageRequest) -> Conversation:
        response = await self.client.apost(
            f"/v1/conversations/{self.id}",
            body=message.model_dump(),
            timeout=30,
        )
        self._update_w_response(response)
        return self

    def iter_messages(self) -> Iterator[MessageEvent]:
        resp = self.client.event_stream(
            f"/v1/conversations/{self.id}/events" + ("?etag=" + self.etag if self.etag else ""),
        )
        for event in resp:
            message = MessageEvent.model_validate_json(event.data)
            if message.role == "user":
                continue
            if message.type == "conversation_end":
                yield message
                break
            yield message
            self.refresh()

    async def aiter_messages(self) -> AsyncIterator[MessageEvent]:
        resp = self.client.aevent_stream(
            f"/v1/conversations/{self.id}/events" + ("?etag=" + self.etag if self.etag else ""),
        )
        async for event in resp:
            message = MessageEvent.model_validate_json(event.data)
            if message.role == "user":
                continue
            if message.type == "conversation_end":
                yield message
                break
            yield message
            await self.arefresh()

    def add_message(self, message: MessageRequest) -> Message:
        response = self.client.post(
            f"/v1/conversations/{self.id}",
            body=message.model_dump(),
            timeout=30,
        )
        self._update_w_response(response)

        if len(self.messages) == 0:
            raise ValueError("No response from the server")

        return Message.model_validate(self.messages[-1])


class MessageEvent(BaseModel):
    role: str
    type: str
    content: str | None = None


class ConversationService(BaseService[Conversation]):
    async def alist(self, page: int = 1, page_size: int = 10) -> list[Conversation]:
        conversations = await self.client.apget(
            "/v1/conversations",
            params={"page": page, "page_size": page_size},
            page=page,
            page_size=page_size,
        )
        return [Conversation(client=self.client, **conversation) for conversation in conversations]

    def list(self, page: int = 1, page_size: int = 10) -> list[Conversation]:
        conversations = self.client.pget(
            "/v1/conversations",
            params={"page": page, "page_size": page_size},
            page=page,
            page_size=page_size,
        )
        return [Conversation(client=self.client, **conversation) for conversation in conversations]

    def create(
        self,
        name: str,
        settings: ConversationSettings | None = None,
        agent_id: str | None = None,
    ) -> Conversation:
        if (settings is None and agent_id is None) or (settings is not None and agent_id is not None):
            raise ValueError("Exactly one of settings or agent_id must be provided")

        params = {}
        if agent_id:
            params["assistant_id"] = agent_id

        # Match CreateConversation schema
        req = {"name": name, "settings": settings.model_dump() if settings else None}

        result = self.client.post(
            "/v1/conversations",
            body=req,
            params=params,
        )
        return Conversation(client=self.client, **result)

    async def acreate(
        self,
        name: str,
        settings: ConversationSettings | None = None,
        agent_id: str | None = None,
    ) -> Conversation:
        if (settings is None and agent_id is None) or (settings is not None and agent_id is not None):
            raise ValueError("Exactly one of settings or agent_id must be provided")

        params = {}
        if agent_id:
            params["assistant_id"] = agent_id

        # Match CreateConversation schema
        req = {"name": name, "settings": settings.model_dump() if settings else None}

        result = await self.client.apost(
            "/v1/conversations",
            body=req,
            params=params,
        )
        return Conversation(client=self.client, **result)

    def get(self, conversation_id: str) -> Conversation:
        result = self.client.get(f"/v1/conversations/{conversation_id}")
        return Conversation(client=self.client, **result)

    async def aget(self, conversation_id: str) -> Conversation:
        result = await self.client.aget(f"/v1/conversations/{conversation_id}")
        return Conversation(client=self.client, **result)

    def update(
        self,
        conversation_id: str,
        name: str | None = None,
        settings: ConversationSettings | None = None,
    ) -> Conversation:
        result = self.client.patch(
            f"/v1/conversations/{conversation_id}",
            {"name": name, "settings": settings.model_dump() if settings else None},
        )
        return Conversation(client=self.client, **result)

    async def aupdate(
        self,
        conversation_id: str,
        name: str | None = None,
        settings: ConversationSettings | None = None,
    ) -> Conversation:
        result = await self.client.apatch(
            f"/v1/conversations/{conversation_id}",
            {"name": name, "settings": settings.model_dump() if settings else None},
        )
        return Conversation(client=self.client, **result)

    def delete(self, conversation_id: str) -> None:
        self.client.delete(f"/v1/conversations/{conversation_id}")

    async def adelete(self, conversation_id: str) -> None:
        await self.client.adelete(f"/v1/conversations/{conversation_id}")



"""Plugin interfaces and abstract base classes"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from noxus_sdk.plugins.manifest import PluginManifest


class PluginSource(ABC):
    """Abstract base class for plugin sources"""

    @abstractmethod
    def get_name(self) -> str:
        """Get the name/identifier of the plugin"""

    @abstractmethod
    async def download_plugin(self, output_dir: str | Path | None = None) -> Path:
        """Download or copy the plugin to the output directory"""

    @abstractmethod
    async def get_manifest(self) -> PluginManifest:
        """Get the manifest of the plugin"""



from __future__ import annotations

import asyncio
import os
import time
from typing import TYPE_CHECKING, Any, BinaryIO

import httpx
from httpx_sse import ServerSentEvent, aconnect_sse, connect_sse

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterator

FileContent = BinaryIO | bytes | str
HttpxFile = tuple[str, tuple[str, FileContent, str | None]]
RequestFiles = dict[str, Any] | list[HttpxFile] | None


class RequestFailedError(Exception):
    pass


class Requester:
    base_url = os.environ.get("NOXUS_BACKEND_URL", "https://backend.noxus.ai")

    def __init__(self, api_key: str, extra_headers: dict | None = None) -> None:
        self.api_key = api_key
        self.extra_headers = extra_headers

    async def _arequest(
        self,
        method: str,
        url: str,
        headers: dict | None = None,
        json: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> httpx.Response:
        headers_ = {"X-API-Key": self.api_key}
        if headers:
            headers_.update(headers)
        if self.extra_headers:
            headers_.update(self.extra_headers)
        ratelimited = True
        while ratelimited:
            async with httpx.AsyncClient() as client:
                response = await client.request(
                    method,
                    f"{self.base_url}{url}",
                    headers=headers_,
                    follow_redirects=True,
                    json=json,
                    files=files,
                    params=params,
                    timeout=timeout or 120,
                )
                if response.status_code == 429:
                    await asyncio.sleep(1)
                    continue
                ratelimited = False
                response.raise_for_status()
                return response
        raise RequestFailedError("Request failed")

    async def arequest(
        self,
        method: str,
        url: str,
        headers: dict | None = None,
        json: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> dict:
        return (
            await self._arequest(
                method,
                url,
                headers=headers,
                json=json,
                files=files,
                params=params,
                timeout=timeout,
            )
        ).json()

    async def aget(
        self,
        url: str,
        headers: dict | None = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> Any:  # noqa: ANN401
        return await self.arequest(
            "GET",
            url,
            headers=headers,
            params=params,
            timeout=timeout,
        )

    async def apget(
        self,
        url: str,
        headers: dict | None = None,
        params: dict | None = None,
        page: int = 1,
        page_size: int = 10,
        timeout: int | None = None,
    ) -> list[dict]:
        params_ = params or {}
        params_["page"] = params_.get("page", page)
        params_["size"] = params_.get("page_size", page_size)

        headers_ = {"X-API-Key": self.api_key}
        if headers:
            headers_.update(headers)
        if self.extra_headers:
            headers_.update(self.extra_headers)
        result = await self.arequest(
            "GET",
            url,
            headers=headers_,
            params=params_,
            timeout=timeout,
        )
        if "items" not in result:
            return []
        return result["items"]

    async def apost(
        self,
        url: str,
        body: Any | None = None,  # noqa: ANN401
        headers: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> dict:
        return await self.arequest(
            "POST",
            url,
            json=body,
            headers=headers,
            files=files,
            params=params,
            timeout=timeout,
        )

    async def apatch(
        self,
        url: str,
        body: Any,  # noqa: ANN401
        headers: dict | None = None,
        timeout: int | None = None,
        params: dict | None = None,
    ) -> dict:
        return await self.arequest(
            "PATCH",
            url,
            json=body,
            headers=headers,
            timeout=timeout,
            params=params,
        )

    async def adelete(
        self,
        url: str,
        headers: dict | None = None,
        timeout: int | None = None,
    ) -> dict:
        return await self.arequest("DELETE", url, headers=headers, timeout=timeout)

    def _request(
        self,
        method: str,
        url: str,
        headers: dict | None = None,
        json: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> httpx.Response:
        headers_ = {"X-API-Key": self.api_key}
        if headers:
            headers_.update(headers)
        if self.extra_headers:
            headers_.update(self.extra_headers)
        ratelimited = True
        while ratelimited:
            response = httpx.request(
                method,
                f"{self.base_url}{url}",
                headers=headers_,
                follow_redirects=True,
                json=json,
                files=files,
                params=params,
                timeout=timeout or 120,
            )
            if response.status_code == 429:
                time.sleep(1)
                continue
            ratelimited = False
            response.raise_for_status()
            return response
        raise RequestFailedError("Request failed")

    def request(
        self,
        method: str,
        url: str,
        headers: dict | None = None,
        json: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> dict:
        response = self._request(
            method,
            url,
            headers=headers,
            json=json,
            files=files,
            params=params,
            timeout=timeout,
        )
        return response.json()

    def event_stream(
        self,
        url: str,
        headers: dict | None = None,
        json: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> Iterator[ServerSentEvent]:
        headers_ = {"X-API-Key": self.api_key}
        if headers:
            headers_.update(headers)
        if self.extra_headers:
            headers_.update(self.extra_headers)
        ratelimited = True
        with httpx.Client() as client:
            while ratelimited:
                with connect_sse(
                    client=client,
                    method="GET",
                    url=f"{self.base_url}{url}",
                    headers=headers_,
                    follow_redirects=True,
                    json=json,
                    files=files,
                    params=params,
                    timeout=timeout or 120,
                ) as response:
                    ratelimited = False
                    yield from response.iter_sse()
        raise RequestFailedError("Request failed")

    async def aevent_stream(
        self,
        url: str,
        headers: dict | None = None,
        json: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> AsyncIterator[ServerSentEvent]:
        headers_ = {"X-API-Key": self.api_key}
        if headers:
            headers_.update(headers)
        if self.extra_headers:
            headers_.update(self.extra_headers)
        ratelimited = True
        async with httpx.AsyncClient() as client:
            while ratelimited:
                async with aconnect_sse(
                    client=client,
                    method="GET",
                    url=f"{self.base_url}{url}",
                    headers=headers_,
                    follow_redirects=True,
                    json=json,
                    files=files,
                    params=params,
                    timeout=timeout or 120,
                ) as response:
                    ratelimited = False
                    async for event in response.aiter_sse():
                        yield event
        raise RequestFailedError("Request failed")

    def get(
        self,
        url: str,
        headers: dict | None = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> Any:  # noqa: ANN401
        return self.request("GET", url, headers=headers, params=params, timeout=timeout)

    def pget(
        self,
        url: str,
        headers: dict | None = None,
        params: dict | None = None,
        page: int = 1,
        page_size: int = 10,
        timeout: int | None = None,
    ) -> list[dict]:
        params_ = params or {}
        params_["page"] = params_.get("page", page)
        params_["size"] = params_.get("page_size", page_size)

        headers_ = {"X-API-Key": self.api_key}
        if headers:
            headers_.update(headers)
        if self.extra_headers:
            headers_.update(self.extra_headers)
        result = self.request(
            "GET",
            url,
            headers=headers_,
            params=params_,
            timeout=timeout,
        )
        if "items" not in result:
            return []
        return result["items"]

    def patch(
        self,
        url: str,
        body: Any,  # noqa: ANN401
        headers: dict | None = None,
        timeout: int | None = None,
        params: dict | None = None,
    ) -> Any:  # noqa: ANN401
        return self.request(
            "PATCH",
            url,
            json=body,
            headers=headers,
            timeout=timeout,
            params=params,
        )

    def post(
        self,
        url: str,
        body: Any | None = None,  # noqa: ANN401
        headers: dict | None = None,
        files: RequestFiles = None,
        params: dict | None = None,
        timeout: int | None = None,
    ) -> Any:  # noqa: ANN401
        return self.request(
            "POST",
            url,
            json=body,
            headers=headers,
            files=files,
            params=params,
            timeout=timeout,
        )

    def delete(
        self,
        url: str,
        headers: dict | None = None,
        timeout: int | None = None,
    ) -> Any:  # noqa: ANN401
        return self.request("DELETE", url, headers=headers, timeout=timeout)


class Client(Requester):
    def __init__(
        self,
        api_key: str,
        base_url: str = "https://backend.noxus.ai",
        extra_headers: dict | None = None,
        *,
        load_nodes: bool = True,
        load_me: bool = True,
    ) -> None:
        from noxus_sdk.resources.admin import AdminService
        from noxus_sdk.resources.agentflows import AgentFlowService
        from noxus_sdk.resources.assistants import AgentService
        from noxus_sdk.resources.conversations import ConversationService
        from noxus_sdk.resources.files import FileService
        from noxus_sdk.resources.knowledge_bases import KnowledgeBaseService
        from noxus_sdk.resources.runs import RunService
        from noxus_sdk.resources.workflows import WorkflowService
        from noxus_sdk.workflows import load_node_types

        self.api_key = api_key
        self.base_url = os.environ.get("NOXUS_BACKEND_URL", base_url)
        self.extra_headers = extra_headers

        if load_nodes:
            self.nodes = self.get_nodes()
            load_node_types(self.nodes)
        else:
            self.nodes = []

        self.workflows = WorkflowService(self)
        self.agentflows = AgentFlowService(self)
        self.agents = AgentService(self)
        self.conversations = ConversationService(self)
        self.knowledge_bases = KnowledgeBaseService(self)
        self.runs = RunService(self)
        self.admin = AdminService(self, enabled=bool(not load_me))
        self.files = FileService(self)
        if load_me:
            self.admin.enabled = self.admin.get_me().tenant_admin

    def get_nodes(self) -> list[dict]:
        return self.get("/v1/nodes")

    async def aget_nodes(self) -> list[dict]:
        return await self.aget("/v1/nodes")

    def get_models(self) -> list[dict]:
        return self.get("/v1/models/llms")

    async def aget_models(self) -> list[dict]:
        return await self.aget("/v1/models/llms")

    def get_chat_presets(self) -> list[dict]:
        return self.get("/v1/models/llms/presets")

    async def aget_chat_presets(self) -> list[dict]:
        return await self.aget("/v1/models/llms/presets")



"""Parameter definition utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from pydantic import Field

if TYPE_CHECKING:
    from noxus_sdk.ncl.displays import BaseConfigDisplay
    from noxus_sdk.ncl.rules import BaseRule

# Constants
ADVANCED_ACCORDION = "Advanced options"
CONFIGURATION_TAB: Literal["Configuration", "Safety", "Human", "Model"] = "Configuration"
MODEL_TAB: Literal["Configuration", "Safety", "Human", "Model"] = "Model"
SAFETY_TAB: Literal["Configuration", "Safety", "Human", "Model"] = "Safety"
HUMAN_TAB: Literal["Configuration", "Safety", "Human", "Model"] = "Human"


def Parameter(  # noqa: N802 - Syntax sugar
    *args,
    display: BaseConfigDisplay | None = None,
    visible: bool = True,
    rules: list[BaseRule] | None = None,
    advanced: bool = False,
    tab: Literal["Configuration", "Safety", "Human", "Model"] = "Configuration",
    accordion: str | None = None,
    optional: bool = False,
    tooltip: str | None = None,
    placeholder: Any | None = None,  # noqa: ANN401 - No way around this for now
    col_span: int | None = None,
    col_justify: Literal["start", "center", "end"] | None = None,
    col_align: Literal["start", "center", "end"] | None = None,
    **kwargs,
) -> Field:
    """Create a configuration parameter with UI metadata."""
    js_field: dict[str, Any] = {}
    if display:
        js_field["display"] = display
    if not visible:
        js_field["visible"] = False
    if rules:
        js_field["rules"] = rules
    if advanced:
        js_field["advanced"] = True
    if optional:
        js_field["optional"] = True
    if tab:
        js_field["tab"] = tab
    if accordion:
        js_field["accordion"] = accordion
    if tooltip:
        js_field["tooltip"] = tooltip
    if placeholder:
        js_field["placeholder"] = placeholder
    if col_span:
        js_field["col_span"] = col_span
    if col_justify:
        js_field["col_justify"] = col_justify
    if col_align:
        js_field["col_align"] = col_align
    return Field(
        *args,
        json_schema_extra=js_field,
        **kwargs,
    )



"""Configuration validation and behavior rules."""

from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel


class BaseRule(BaseModel):
    """Base class for configuration rules."""

    type: Literal["base_rule"]


class TypeSetBy(BaseRule):
    type: Literal["type_set_by"] = "type_set_by"  # type: ignore
    config_source: str


class KeysSetBy(BaseRule):
    type: Literal["keys_set_by"] = "keys_set_by"  # type: ignore
    config_source: str


class VisibleIfGroup(BaseRule):
    type: Literal["visible_if_group"] = "visible_if_group"  # type: ignore
    conds: list[VisibleIf | VisibleIfGroup]
    mode: Literal["and", "or"] = "and"


class VisibleIf(BaseRule):
    type: Literal["visible_if"] = "visible_if"  # type: ignore
    config_source: str
    value: Any


class DisabledIf(BaseRule):
    type: Literal["disabled_if"] = "disabled_if"  # type: ignore
    config_source: str
    value: Any


class SetsConnectorType(BaseRule):
    type: Literal["sets_connector_type"] = "sets_connector_type"  # type: ignore
    connector: str


class SetsConnectorKeys(BaseRule):
    type: Literal["sets_connector_keys"] = "sets_connector_keys"  # type: ignore
    connector: str


class SetsConnectorTypeAndKeys(BaseRule):
    type: Literal["sets_connector_type_and_keys"] = "sets_connector_type_and_keys"  # type: ignore
    connector: str


class SetsConnectorOrConfigVisibility(BaseRule):
    type: Literal["sets_connector_or_config_visibility"] = "sets_connector_or_config_visibility"  # type: ignore


class TypeAndKeysSetBy(BaseRule):
    type: Literal["type_and_keys_set_by"] = "type_and_keys_set_by"  # type: ignore
    config_source: str


# Union type for all rules
AnyRule = (
    TypeSetBy
    | VisibleIf
    | VisibleIfGroup
    | DisabledIf
    | SetsConnectorType
    | SetsConnectorKeys
    | SetsConnectorTypeAndKeys
    | TypeAndKeysSetBy
    | SetsConnectorOrConfigVisibility
)



"""Common error classes for noxus-sdk"""


class NoxusError(Exception):
    """Base exception for all noxus-sdk errors"""


class NodeDefinitionError(NoxusError):
    """Error in node definition"""


class UnexpectedError(NoxusError):
    """Unexpected error during execution"""


class PluginValidationError(NoxusError):
    """Plugin validation failed"""



"""Base plugin class for plugin development"""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Literal, TypeVar, get_args

from pydantic import BaseModel

if TYPE_CHECKING:
    from noxus_sdk.integrations.base import BaseIntegration
    from noxus_sdk.nodes.base import BaseNode

from noxus_sdk.ncl import serialize_config
from noxus_sdk.plugins.manifest import PluginManifest
from noxus_sdk.plugins.types import PluginCategory
from noxus_sdk.schemas import ValidationResult


class PluginConfiguration(BaseModel):
    """Configuration for the plugin"""

    @classmethod
    def serialize(cls) -> dict:
        """Get the serialized configuration for the plugin"""

        return serialize_config(cls)

    # Can be overridden to specify custom validation
    def validate_config(self) -> ValidationResult:
        """Validate the configuration for the plugin"""
        return ValidationResult(valid=True)


ConfigType = TypeVar("ConfigType", bound=PluginConfiguration)


class BasePlugin(Generic[ConfigType]):
    """Base class for plugin development."""

    # Core plugin metadata
    name: str  # Unique identifier used for plugin lookup, dependencies, and database storage
    display_name: str  # Human-readable name for display purposes
    version: str  # Semantic versioning (e.g. 1.0.0)
    description: str  # Short description of the plugin
    category: PluginCategory = PluginCategory.OTHER  # Category of the plugin
    author: str  # Author of the plugin

    # Execution configuration
    execution: Literal["runtime", "docker", "remote"] = "runtime"

    # Required for execution == "docker"
    image: str | None = None

    # Required for execution == "remote"
    endpoint: str | None = None

    # Internal variables (not exposed to the user)
    _config_class: type[ConfigType]  # Used for internal purposes like getting the configuration class

    def __init_subclass__(cls) -> None:
        """Set the configuration class for the plugin when the sublcass is created"""
        # get_args(cls.__orig_bases__[0])[0] -> PluginConfiguration class defined by the user
        cls._config_class = get_args(cls.__orig_bases__[0])[0]  # type: ignore
        return super().__init_subclass__()

    @classmethod
    def get_config_class(cls) -> type[ConfigType]:
        return cls._config_class

    @classmethod
    def get_manifest(cls) -> PluginManifest:
        """Get the manifest for the plugin"""

        plugin_instance = cls()
        provided_nodes = plugin_instance.nodes()
        provided_integrations = plugin_instance.integrations()

        # Plugin must provide at least one node OR integration
        if not provided_nodes and not provided_integrations:
            raise ValueError(
                f"Plugin '{cls.name}' must provide at least one node or integration",
            )

        return PluginManifest(
            name=cls.name,
            display_name=cls.display_name,
            version=cls.version,
            description=cls.description,
            category=cls.category,
            author=cls.author,
            config=cls.get_config_class().serialize(),
            execution=cls.execution,
            image=cls.image,
            endpoint=cls.endpoint,
            nodes=[node.get_definition() for node in provided_nodes],
            integrations=[integration.get_definition() for integration in provided_integrations],
        )

    # Methods to override

    def nodes(self) -> list[type[BaseNode]]:
        """Return list of node classes provided by this plugin"""
        return []

    def integrations(self) -> list[type[BaseIntegration]]:
        """Return list of integration classes provided by this plugin"""
        return []



from noxus_sdk.workflows.agentflow import AgentFlowDefinition
from noxus_sdk.workflows.workflow import (
    ConfigError,
    WorkflowDefinition,
    load_node_types,
)

__all__ = [
    "AgentFlowDefinition",
    "WorkflowDefinition",
    "ConfigError",
    "load_node_types",
]



import typer

from noxus_sdk import __version__
from noxus_sdk.cli.commands.plugin import app as plugin_app

app = typer.Typer(
    help="Noxus CLI - Software development kit to extend the Noxus platform",
)

app.add_typer(plugin_app, name="plugin")


@app.command()
def version() -> None:
    """Show version information"""
    typer.echo(f"noxus version {__version__}")


def main() -> None:
    """Main entry point for the CLI"""
    app()


if __name__ == "__main__":
    main()



from noxus_sdk.resources.assistants import *  # noqa: F403
from noxus_sdk.resources.conversations import *  # noqa: F403
from noxus_sdk.resources.knowledge_bases import *  # noqa: F403
from noxus_sdk.resources.workflows import *  # noqa: F403



from __future__ import annotations

import asyncio
import time
from typing import TYPE_CHECKING

from pydantic import ConfigDict

from noxus_sdk.resources.base import BaseResource, BaseService

if TYPE_CHECKING:
    import builtins


class RunFailureError(Exception):
    pass


class Run(BaseResource):
    model_config = ConfigDict(validate_assignment=True)

    id: str
    group_id: str
    workflow_id: str
    input: dict
    node_ids: list[str] | None = None
    status: str
    progress: int
    progress_details: dict | None = None
    created_at: str
    finished_at: str | None = None
    output: dict | None = None
    workflow_definition: dict | None = None

    def refresh(self) -> Run:
        response = self.client.get(f"/v1/workflows/{self.workflow_id}/runs/{self.id}")
        for key, value in response.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    async def arefresh(self) -> Run:
        response = await self.client.aget(
            f"/v1/workflows/{self.workflow_id}/runs/{self.id}",
        )
        for key, value in response.items():
            if hasattr(self, key):
                setattr(self, key, value)
        return self

    def wait(
        self,
        interval: int = 5,
        *,
        output_only: bool = False,
    ) -> Run | dict | None:
        while self.status not in ["failed", "completed", "awaiting_human_feedback"]:
            time.sleep(interval)
            self.refresh()

        if self.status == "failed":
            raise RunFailureError(self.status)

        if output_only:
            return self.output
        return self

    async def a_wait(
        self,
        interval: int = 5,
        *,
        output_only: bool = False,
    ) -> Run | dict | None:
        while self.status not in ["failed", "completed", "awaiting_human_feedback"]:
            await asyncio.sleep(interval)
            await self.arefresh()

        if self.status == "failed":
            raise RunFailureError(self.status)

        if output_only:
            return self.output
        return self

    def get_status(self) -> str:
        return self.status


class RunService(BaseService[Run]):
    def get(self, workflow_id: str, run_id: str) -> Run:
        response = self.client.get(f"/v1/workflows/{workflow_id}/run/{run_id}")
        return Run(client=self.client, **response)

    async def aget(self, workflow_id: str, run_id: str) -> Run:
        response = await self.client.aget(f"/v1/workflows/{workflow_id}/run/{run_id}")
        return Run(client=self.client, **response)

    def list(self, workflow_id: str, page: int = 1, page_size: int = 10) -> list[Run]:
        response = self.client.pget(
            f"/v1/workflows/{workflow_id}/runs",
            params={"page": page, "page_size": page_size},
        )
        return [Run(client=self.client, **run) for run in response]

    async def alist(
        self,
        workflow_id: str,
        page: int = 1,
        page_size: int = 10,
    ) -> builtins.list[Run]:
        response = await self.client.apget(
            f"/v1/workflows/{workflow_id}/runs",
            params={"page": page, "page_size": page_size},
            page=page,
            page_size=page_size,
        )
        return [Run(client=self.client, **run) for run in response]



from __future__ import annotations

import enum
import uuid
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, Field, TypeAdapter, model_validator
from pydantic.config import ConfigDict

from noxus_sdk.client import Client

if TYPE_CHECKING:
    from noxus_sdk.resources.runs import Run
    from noxus_sdk.resources.workflows import WorkflowVersion


class ConfigError(Exception):
    pass


DATA_TYPE_MAP = {
    "int": int,
    "float": (float, int),
    "bool": bool,
    "dict": dict,
    "str": str,
    "list": list,
}


class DataType(str, enum.Enum):
    int = "int"
    float = "float"
    bool = "bool"
    dict = "dict"
    str = "str"
    list = "list"
    Image = "Image"
    Audio = "Audio"
    File = "File"
    Quote = "Quote"
    Custom = "Custom"
    NoxusAny = "Any"
    GoogleSheet = "GoogleSheet"
    SourceType = "SourceType"

    def validate(self, value: Any) -> Any:  # noqa: ANN401 - This is a valid use case for Any
        if self.value in DATA_TYPE_MAP and not isinstance(
            value,
            DATA_TYPE_MAP[self.value],
        ):
            raise TypeError(
                f"value '{value}' for data type {self} is invalid",
            )

        return value


class ConfigDefinition(BaseModel):
    type: DataType
    description: str | None
    visible: bool
    optional: bool
    default: Any

    def check_value(self, key: str, value: Any) -> None:  # noqa: ANN401 - This is a valid use case for Any
        if value is None:
            if not self.optional:
                raise ConfigError(f"Missing required config value for {key}")
            if self.default is not None:
                value = self.default
            if value == self.default:
                return
        try:
            self.type.validate(value)
        except Exception as e:
            raise ConfigError(f"Invalid config value for {key}: [{e}]") from e


class NodeDefinition(BaseModel):
    type: str
    title: str
    description: str
    integrations: list[str]
    inputs: list[dict]
    outputs: list[dict]
    config: dict[str, ConfigDefinition]
    is_available: bool
    visible: bool
    config_endpoint: str | None


NODE_TYPES: dict[str, NodeDefinition] = {}


def load_node_types(nodes_: list[dict]) -> None:
    NODE_TYPES.clear()
    nodes: list[NodeDefinition] = TypeAdapter(list[NodeDefinition]).validate_python(
        nodes_,
    )
    for node in nodes:
        NODE_TYPES[node.type] = node


class ConnectorType(str, enum.Enum):
    variable_connector = "variable_connector"
    variable_type_connector = "variable_type_connector"
    variable_type_size_connector = "variable_type_size_connector"
    variable_type_input = "variable_type_input"
    variable_type_output = "variable_type_output"
    connector = "connector"
    input = "input"
    output = "output"


class NodeInput(BaseModel):
    node_id: str
    name: str
    fixed_value: Any = None
    type: ConnectorType

    @property
    def id(self) -> str:
        return f"{self.node_id}::{self.name}"


class EdgePoint(BaseModel):
    node_id: str
    connector_name: str
    key: str | None = None
    optional: bool = False


class Edge(BaseModel):
    from_id: EdgePoint
    to_id: EdgePoint
    id: str | None = None


class NodeOutput(BaseModel):
    node_id: str
    name: str
    type: ConnectorType

    @property
    def id(self) -> str:
        return f"{self.node_id}::{self.name}"


class Node(BaseModel):
    type: str
    id: str
    name: str = ""
    display: dict = {}

    node_config: dict = {}
    connector_config: dict = {}
    config_definition: dict[str, ConfigDefinition] = {}
    subflow_config: dict | None = None
    subflow_id: str | None = None
    inputs: list[NodeInput] = []
    outputs: list[NodeOutput] = []

    def input(
        self,
        name: str | None = None,
        key: str | None = None,
        type_definition: str | None = None,
        *,
        type_definition_is_list: bool = False,
    ) -> EdgePoint:
        if name is None:
            if len(self.inputs) != 1:
                raise ValueError("Multiple inputs found, please specify a name")
            name = self.inputs[0].name
        i = {i.name: i for i in self.inputs}
        if name not in i:
            raise KeyError(f"Input {name} not found (possible: {list(i.keys())})")
        _input = i[name]
        if _input.type in ["variable_connector", "variable_type_size_connector"]:
            if key is None:
                raise ValueError("key is required for variable_connector")
            connector_config = self.connector_config.get("inputs")
            if not connector_config:
                raise ValueError(f"connector_config is missing {self.connector_config}")
            connector_inputs = {i["name"]: i for i in connector_config}
            if name not in connector_inputs:
                raise ValueError(
                    f"Invalid key: {name} (possible: {list(connector_inputs.keys())})",
                )
            connector_keys = connector_inputs[name].get("keys", [])
            if key not in connector_keys:
                connector_keys.append(key)
            if _input.type == "variable_type_size_connector":
                if type_definition is not None:
                    raise ValueError(
                        f"type_definition is required for variable_type_size_connector ({self.type}.{name})",
                    )
                type_definitions = connector_inputs[name].get("type_definitions", {})
                choices = connector_inputs[name].get("choices", [])
                if key not in type_definitions:
                    typedef = {
                        "data_type": type_definition,
                        "is_list": type_definition_is_list,
                    }
                    type_definitions[key] = typedef
                    if type_definition not in [c["data_type"] for c in choices]:
                        choices.append(typedef)
                    connector_inputs[name]["choices"] = choices
                    connector_inputs[name]["type_definitions"] = type_definitions
            return EdgePoint(
                node_id=_input.node_id,
                connector_name=_input.name,
                key=key,
            )
        return EdgePoint(node_id=_input.node_id, connector_name=_input.name, key=None)

    def output(
        self,
        name: str | None = None,
        key: str | None = None,
        type_definition: str | None = None,
        *,
        type_definition_is_list: bool = False,
    ) -> EdgePoint:
        if name is None:
            if len(self.outputs) != 1:
                raise ValueError("Multiple outputs found, please specify a name")
            name = self.outputs[0].name
        i = {i.name: i for i in self.outputs}
        if name not in i:
            raise KeyError(f"Output {name} not found (possible: {list(i.keys())})")
        output = i[name]
        if output.type in ["variable_connector", "variable_type_size_connector"]:
            if key is None:
                raise ValueError("key is required for variable_connector")
            connector_config = self.connector_config.get("outputs")
            if not connector_config:
                raise ValueError("connector_config is missing")
            connector_outputs = {o["name"]: o for o in connector_config}
            if name not in connector_outputs:
                raise ValueError(
                    f"Invalid key: {name} (possible: {list(connector_outputs.keys())})",
                )
            connector_keys = connector_outputs[name].get("keys", [])
            if key not in connector_keys:
                connector_keys.append(key)
            if output.type == "variable_type_size_connector":
                if type_definition is not None:
                    raise ValueError(
                        f"type_definition is required for variable_type_size_connector ({self.type}.{name})",
                    )
                type_definitions = connector_outputs[name].get("type_definitions", {})
                choices = connector_outputs[name].get("choices", [])
                if key not in type_definitions:
                    typedef = {
                        "data_type": type_definition,
                        "is_list": type_definition_is_list,
                    }
                    type_definitions[key] = typedef
                    if type_definition not in [c["data_type"] for c in choices]:
                        choices.append(typedef)
                    connector_outputs[name]["type_definitions"] = type_definitions
                    connector_outputs[name]["choices"] = choices
            return EdgePoint(
                node_id=output.node_id,
                connector_name=output.name,
                key=key,
            )
        return EdgePoint(node_id=output.node_id, connector_name=output.name, key=None)

    def create(self, x: int, y: int) -> Node:
        node_type = NODE_TYPES.get(self.type)
        if not node_type:
            raise ValueError(f"Node type {self.type} not found")
        self.config_definition = node_type.config
        self.inputs = [NodeInput(node_id=str(self.id), name=i["name"], type=i["type"]) for i in node_type.inputs]
        self.outputs = [
            NodeOutput(node_id=str(self.id), name=output["name"], type=output["type"]) for output in node_type.outputs
        ]
        if not self.connector_config:
            self.connector_config = {
                "inputs": node_type.inputs,
                "outputs": node_type.outputs,
            }
        self.name = node_type.title
        self.display = {"position": {"x": x, "y": y}}
        return self

    def config(self, **kwargs) -> Node:
        for key, value in kwargs.items():
            if key not in self.config_definition:
                raise ConfigError(
                    f"Invalid config key: {key} (possible: {[k for k,v in self.config_definition.items() if v.visible]})",
                )
            self.config_definition[key].check_value(key, value)
            self.node_config[key] = value
        for k, v in self.config_definition.items():
            if k not in self.node_config:
                v.check_value(k, None)
        return self

    @model_validator(mode="after")
    def call_create_after_validate(self) -> Node:
        old_display = dict(self.display)
        self.create(0, 0)
        if old_display:
            self.display = old_display
        return self


class WorkflowDefinition(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    client: Client | None = Field(default=None, exclude=True)
    id: str = ""
    group_id: str | None = Field(default=None, exclude=True)
    name: str = "Untitled Workflow"
    type: str = "flow"
    nodes: list[Node] = []
    edges: list[Edge] = []
    x: int = 0
    error_handler: uuid.UUID | None = None

    @model_validator(mode="before")
    @classmethod
    def _definition_flattener(cls, values: dict) -> dict:
        if "definition" in values:
            values["nodes"] = values["definition"]["nodes"]
            values["edges"] = values["definition"]["edges"]
        return values

    def to_noxus(self) -> dict:
        d = {
            "name": self.name,
            "type": self.type,
            "definition": {
                "nodes": [n.model_dump() for n in self.nodes],
                "edges": [e.model_dump() for e in self.edges],
            },
        }
        if self.error_handler:
            d["error_handler"] = str(self.error_handler)
        return d

    def refresh_from_data(
        self,
        client: Client | None = None,
        **data,
    ) -> WorkflowDefinition:
        n = self.__class__.model_validate(data)
        for k in n.model_fields_set:
            v = getattr(n, k)
            setattr(self, k, v)
        self.client = client
        return self

    def refresh(self) -> WorkflowDefinition:
        if not self.client:
            raise ValueError("Client not set")
        response = self.client.get(f"/v1/workflows/{self.id}")
        self.refresh_from_data(client=self.client, **response)
        return self

    async def arefresh(self) -> WorkflowDefinition:
        if not self.client:
            raise ValueError("Client not set")
        response = await self.client.aget(f"/v1/workflows/{self.id}")
        self.refresh_from_data(client=self.client, **response)
        return self

    def run(
        self,
        body: dict[str, Any],
        workflow_version_id: uuid.UUID | str | None = None,
    ) -> Run:
        from noxus_sdk.resources.runs import Run

        if not self.client:
            raise ValueError("Client not set")
        url = f"/v1/workflows/{self.id}/runs"
        req: dict[str, Any] = {"input": body}
        if workflow_version_id:
            req["workflow_version_id"] = str(workflow_version_id)

        response = self.client.post(url, req)
        return Run(client=self.client, **response)

    async def arun(
        self,
        body: dict[str, Any],
        workflow_version_id: uuid.UUID | str | None = None,
    ) -> Run:
        if not self.client:
            raise ValueError("Client not set")
        from noxus_sdk.resources.runs import Run

        req: dict[str, Any] = {"input": body}
        if workflow_version_id:
            req["workflow_version_id"] = str(workflow_version_id)
        response = await self.client.apost(f"/v1/workflows/{self.id}/runs", req)
        return Run(client=self.client, **response)

    def update(self, *, force: bool = False) -> WorkflowDefinition:
        if not self.client:
            raise ValueError("Client not set")
        w = self.client.workflows.update(self.id, self, force=force)
        self.refresh_from_data(client=self.client, **w.model_dump())
        return w

    async def aupdate(self, *, force: bool = False) -> WorkflowDefinition:
        if not self.client:
            raise ValueError("Client not set")
        w = await self.client.workflows.aupdate(self.id, self, force=force)
        self.refresh_from_data(client=self.client, **w.model_dump())
        return w

    def save(self) -> WorkflowDefinition:
        if not self.client:
            raise ValueError("Client not set")
        return self.client.workflows.save(self)

    async def asave(self) -> WorkflowDefinition:
        if not self.client:
            raise ValueError("Client not set")
        return await self.client.workflows.asave(self)

    def save_version(
        self,
        name: str,
        description: str | None = None,
    ) -> WorkflowVersion:
        if not self.client:
            raise ValueError("Client not set")
        return self.client.workflows.save_version(self.id, self, name, description)

    async def asave_version(
        self,
        name: str,
        description: str | None = None,
    ) -> WorkflowVersion:
        if not self.client:
            raise ValueError("Client not set")
        return await self.client.workflows.asave_version(
            self.id,
            self,
            name,
            description,
        )

    def update_version(
        self,
        version_id: str,
        name: str,
        description: str | None,
    ) -> WorkflowVersion:
        if not self.client:
            raise ValueError("Client not set")
        return self.client.workflows.update_version(
            self.id,
            version_id,
            name,
            description,
            self,
        )

    async def aupdate_version(
        self,
        version_id: str,
        name: str,
        description: str | None,
    ) -> WorkflowVersion:
        if not self.client:
            raise ValueError("Client not set")
        return await self.client.workflows.aupdate_version(
            self.id,
            version_id,
            name,
            description,
            self,
        )

    def list_versions(self) -> list[WorkflowVersion]:
        if not self.client:
            raise ValueError("Client not set")
        return self.client.workflows.list_versions(self.id)

    async def alist_versions(self) -> list[WorkflowVersion]:
        if not self.client:
            raise ValueError("Client not set")
        return await self.client.workflows.alist_versions(self.id)

    def verify_name_legal(self, name: str) -> None:
        if name in [
            "AgentStartNode",
            "AgentEndNode",
            "AgentMessageSendNode",
            "ChoiceAgentNode",
            "FormExtractionAgentNode",
            "BasicAgentNode",
            "ChatAgentNode",
        ]:
            raise ValueError(f"Invalid node name: {name}")

    def node(self, name: str) -> Node:
        self.verify_name_legal(name)
        self.x += 350
        n = Node(id=str(uuid.uuid4()), type=name)
        n.create(x=self.x, y=0)
        self.nodes.append(n)
        return n

    def link(self, from_node: EdgePoint, to_node: EdgePoint) -> Edge:
        e = Edge(id=str(uuid.uuid4()), from_id=from_node, to_id=to_node)
        self.edges.append(e)
        return e

    def link_many(self, *nodes: Node) -> None:
        for i in range(len(nodes) - 1):
            if not len(nodes[i].outputs) == 1:
                raise ValueError(
                    f"Expected 1 output for {nodes[i].type}, got {len(nodes[i].outputs)}",
                )

            if nodes[i].outputs[0].type == "variable_connector":
                raise ValueError(
                    f"A key is required for variable_connector output so unable to link {nodes[i].type} to {nodes[i + 1].type} automatically",
                )
            if not len(nodes[i + 1].inputs) == 1:
                raise ValueError(
                    f"Expected 1 input for {nodes[i + 1].type}, got {len(nodes[i + 1].inputs)}",
                )
            if nodes[i + 1].inputs[0].type == "variable_connector":
                raise ValueError(
                    f"A key is required for variable_connector input so unable to link {nodes[i].type} to {nodes[i + 1].type} automatically",
                )
            self.link(nodes[i].output(), nodes[i + 1].input())



"""Output type definitions."""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel

from noxus_sdk.ncl.parameters import Parameter

# ========================================
# BASE CLASS
# ========================================


class OutputTypeBase(BaseModel):
    """Base class for output type definitions."""

    name: str
    required: bool = True


# ========================================
# OUTPUT TYPE IMPLEMENTATIONS
# ========================================


class BoolOutputType(OutputTypeBase):
    type: Literal["checkbox"] = "checkbox"
    initial_value: dict | str | bool | None = None

    true_label: str = Parameter(default="Accept", advanced=True)
    false_label: str = Parameter(default="Reject", advanced=True)


class SelectOutputType(OutputTypeBase):
    type: Literal["select"] = "select"
    initial_value: str | None = None
    options: list[str] = []


class StringOutputType(OutputTypeBase):
    type: Literal["text"] = "text"
    initial_value: dict | str | None = None


class FileOutputType(OutputTypeBase):
    type: Literal["file"] = "file"
    initial_value: dict | str | None = None


# ========================================
# UNION TYPE FOR ALL OUTPUTS
# ========================================

OutputType = BoolOutputType | StringOutputType | FileOutputType | SelectOutputType



"""Upload plugin source for handling uploaded plugin files"""

from __future__ import annotations

import asyncio
import json
import os
import shutil
import tarfile
import tempfile
import zipfile
from pathlib import Path
from typing import Literal

import aiofiles
import httpx
from loguru import logger
from pydantic import BaseModel, Field

from noxus_sdk.plugins.interfaces import PluginSource
from noxus_sdk.plugins.manifest import PluginManifest


def _is_safe_path(base_path: Path, member_path: str) -> bool:
    """Check if the member path is safe to extract (no path traversal)"""
    # Normalize the member path and resolve any .. components
    try:
        full_path = (base_path / member_path).resolve()
        base_path_resolved = base_path.resolve()

        # Check if the resolved path is within the base directory
        return str(full_path).startswith(str(base_path_resolved) + os.sep) or full_path == base_path_resolved
    except (OSError, ValueError):
        # If path resolution fails, consider it unsafe
        return False


def _validate_member_name(name: str) -> bool:
    """Validate archive member name for security"""
    # Check for dangerous patterns
    return not (
        ".." in name or name.startswith(("/", "\\")) or (":" in name and os.name == "nt")  # Windows drive letters
    )


class UploadPluginSource(PluginSource, BaseModel):
    """Upload plugin source for uploaded files"""

    type: Literal["upload"] = "upload"
    file_id: str = Field(..., description="File ID to download from platform")
    filename: str = Field(..., description="Original filename of the uploaded plugin")

    def get_name(self) -> str:
        """Get the name of the plugin from the filename (will be replaced by manifest name)"""
        # Remove .zip extension and any version info
        return Path(self.filename).stem

    async def download_plugin(self, output_dir: str | Path | None = None) -> Path:
        """Download and extract the uploaded plugin file to the output directory"""
        if output_dir is None:
            raise ValueError("Output directory must be specified")

        output_dir = Path(output_dir)

        # Download file from platform API
        temp_file_path = await self._download_file_from_platform()

        try:
            logger.info(f"Extracting plugin code to {output_dir}")

            # Determine file type and extract accordingly
            if self.filename.lower().endswith(".zip"):
                await self._extract_archive(temp_file_path, output_dir, "zip")
            elif self.filename.lower().endswith((".tar.gz", ".tgz", ".tar")):
                await self._extract_archive(temp_file_path, output_dir, "tar")
            else:
                raise ValueError(f"Unsupported file format: {self.filename}")

            return output_dir

        finally:
            # Clean up downloaded temp file
            if temp_file_path.exists():
                temp_file_path.unlink()

    async def _download_file_from_platform(self) -> Path:
        """Download file from platform API using file ID"""
        platform_url = os.getenv("NOXUS_PLATFORM_URL", "http://localhost:8000")
        # Hardcoded internal API key for plugin server authentication
        # TODO(@andre): Replace with proper service-to-service authentication
        api_key = os.getenv("NOXUS_PLATFORM_API_KEY", "internal-plugin-server-key-2025")

        if not api_key:
            raise ValueError("NOXUS_PLATFORM_API_KEY environment variable is required")

        url = f"{platform_url}/plugin-server/files/{self.file_id}"
        headers = {"X-API-Key": api_key}

        logger.info(f"Downloading file {self.file_id} from platform: {url}")

        async with httpx.AsyncClient() as client:
            response = await client.get(url, headers=headers, timeout=60.0)

            if response.status_code == 404:
                raise FileNotFoundError(
                    f"File with ID {self.file_id} not found on platform",
                )
            if response.status_code == 401:
                raise ValueError("Invalid API key for platform access")

            response.raise_for_status()

            # Create temporary file
            temp_file = tempfile.NamedTemporaryFile(
                suffix=f"_{self.filename}",
                delete=False,
            )
            temp_file.write(response.content)
            temp_file.close()

            logger.info(f"Downloaded file to temporary location: {temp_file.name}")
            return Path(temp_file.name)

    async def _extract_archive(
        self,
        source_path: Path,
        target_path: Path,
        archive_type: str,
    ) -> None:
        """Extract an archive file (zip or tar), handling single-folder archives correctly"""
        try:
            # Get file names from archive
            if archive_type == "zip":
                await self._extract_zip_archive(source_path, target_path)
            elif archive_type == "tar":
                await self._extract_tar_archive(source_path, target_path)
            else:
                raise ValueError(f"Unsupported archive type: {archive_type}")

        except (zipfile.BadZipFile, tarfile.TarError) as e:
            raise ValueError(f"Invalid {archive_type} file: {source_path}") from e

    async def _extract_zip_archive(self, source_path: Path, target_path: Path) -> None:
        """Securely extract ZIP archive with path validation"""

        def _sync_extract_zip() -> None:
            with zipfile.ZipFile(source_path, "r") as archive_ref:
                names = archive_ref.namelist()
                self._secure_extract_archive(archive_ref, names, target_path, "ZIP")

        # Run in thread pool to avoid blocking
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, _sync_extract_zip)

    async def _extract_tar_archive(self, source_path: Path, target_path: Path) -> None:
        """Securely extract TAR archive with path validation"""

        def _sync_extract_tar() -> None:
            with tarfile.open(source_path, "r:*") as archive_ref:
                names = archive_ref.getnames()
                self._secure_extract_archive(archive_ref, names, target_path, "TAR")

        # Run in thread pool to avoid blocking
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, _sync_extract_tar)

    def _secure_extract_archive(
        self,
        archive_ref: zipfile.ZipFile | tarfile.TarFile,
        names: list[str],
        target_path: Path,
        archive_type: str,
    ) -> None:
        """Common secure extraction logic for both ZIP and TAR archives"""
        if not names:
            return

        # Filter out unsafe names
        safe_names = []
        for name in names:
            if _validate_member_name(name) and _is_safe_path(target_path, name):
                safe_names.append(name)
            else:
                logger.warning(f"Skipping unsafe path in {archive_type}: {name}")

        if not safe_names:
            logger.warning(f"No safe files found in {archive_type} archive")
            return

        # Handle single folder extraction
        self._handle_single_folder_extraction(archive_ref, safe_names, target_path)

    def _handle_single_folder_extraction(
        self,
        archive_ref: zipfile.ZipFile | tarfile.TarFile,
        safe_names: list[str],
        target_path: Path,
    ) -> None:
        """Handle single folder extraction logic for both ZIP and TAR files"""
        # Get top-level entries (excluding system files)
        top_level_entries = []
        for name in safe_names:
            top_part = name.split("/")[0] if "/" in name else name
            if not top_part.startswith(("__", ".")) and top_part not in top_level_entries:
                top_level_entries.append(top_part)

        # Ensure target directory exists
        target_path.mkdir(parents=True, exist_ok=True)

        # If there's exactly one top-level directory, extract its contents
        if len(top_level_entries) == 1 and any(name.startswith(f"{top_level_entries[0]}/") for name in safe_names):
            root_dir = top_level_entries[0]
            # Extract to temp dir first, then copy contents
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_target = Path(temp_dir)
                # Extract only safe members
                self._extract_members(archive_ref, safe_names, temp_target)

                extracted_root = temp_target / root_dir
                if extracted_root.exists() and any(extracted_root.iterdir()):
                    # Copy contents of the extracted directory to target
                    for item in extracted_root.iterdir():
                        if item.is_dir():
                            shutil.copytree(
                                item,
                                target_path / item.name,
                                dirs_exist_ok=True,
                            )
                        else:
                            shutil.copy2(item, target_path / item.name)
                else:
                    # Fallback: extract directly to target
                    self._extract_members(archive_ref, safe_names, target_path)
        else:
            # Multiple top-level entries or files, extract directly
            self._extract_members(archive_ref, safe_names, target_path)

    def _extract_members(
        self,
        archive_ref: zipfile.ZipFile | tarfile.TarFile,
        safe_names: list[str],
        target_path: Path,
    ) -> None:
        """Extract members from archive, handling both ZIP and TAR formats"""
        if isinstance(archive_ref, zipfile.ZipFile):
            for name in safe_names:
                archive_ref.extract(name, target_path)
        elif isinstance(archive_ref, tarfile.TarFile):
            for name in safe_names:
                member = archive_ref.getmember(name)
                archive_ref.extract(member, target_path)

    async def get_manifest(self) -> PluginManifest:
        """Get the manifest by extracting to a temp directory"""
        with tempfile.TemporaryDirectory() as temp_dir:
            # Extract the plugin to temp directory
            plugin_path = await self.download_plugin(temp_dir)

            # Read manifest.json from extracted plugin
            manifest_file = plugin_path / "manifest.json"
            if not manifest_file.exists():
                # Show all files and directories for better debugging
                available_items = []
                for item in plugin_path.glob("*"):
                    if item.is_dir():
                        available_items.append(f"{item.name}/")
                    else:
                        available_items.append(item.name)

                logger.error(
                    f"manifest.json not found in {plugin_path}. Available items: {available_items}",
                )

                # Also check if manifest.json exists in any subdirectories
                for subdir in plugin_path.glob("*/"):
                    sub_manifest = subdir / "manifest.json"
                    if sub_manifest.exists():
                        logger.error(f"Found manifest.json in subdirectory: {subdir}")

                raise FileNotFoundError("manifest.json not found in plugin")

            # Load and parse manifest
            async with aiofiles.open(manifest_file, encoding="utf-8") as f:
                content = await f.read()
                manifest_data = json.loads(content)

            return PluginManifest(**manifest_data)



from __future__ import annotations

from pydantic import BaseModel

from noxus_sdk.integrations.types import IntegrationType


class IntegrationDefinition(BaseModel):
    """Definition schema for integrations in plugin manifests"""

    name: str
    type: IntegrationType
    display_name: str
    description: str | None = None
    image: str
    scopes: list[str] | None = None

    properties: dict = {}  # Used to add extra properties to the integration definition



from noxus_sdk.ncl import ConfigSelect, Parameter
from noxus_sdk.nodes import (
    BaseNode,
    Connector,
    DataType,
    NodeCategory,
    NodeConfiguration,
    TypeDefinition,
)
from noxus_sdk.plugins import RemoteExecutionContext


class ExampleNodeConfiguration(NodeConfiguration):
    """Configuration for the Example node that sums two numbers."""

    operation: str = Parameter(default="", display=ConfigSelect(label="Operation", values=["Add", "Subtract", "Multiply", "Divide"]))


class ExampleNode(BaseNode[ExampleNodeConfiguration]):
    """{{ cookiecutter.description }}"""

    inputs = [
        Connector(name="a", label="A", definition=TypeDefinition(data_type=DataType.str)),
        Connector(name="b", label="B", definition=TypeDefinition(data_type=DataType.str)),
    ]
    outputs = [
        Connector(name="result", label="Sum", definition=TypeDefinition(data_type=DataType.str)),
    ]

    node_name = "ExampleNode"
    title = "Example node"
    description = "An example node that performs an operation on two numbers."
    category = NodeCategory.OTHER

    async def call(self, ctx: RemoteExecutionContext, a: str, b: str):
        """Return the sum of configuration values a and b as a string."""
        a_int, b_int = int(a), int(b)

        if self.config.operation == "Add":
            total = a_int + b_int
        elif self.config.operation == "Subtract":
            total = a_int - b_int
        elif self.config.operation == "Multiply":
            total = a_int * b_int
        elif self.config.operation == "Divide":
            total = a_int / b_int
        else:
            total = 0

        return {"result": str(total)}



import base64
from uuid import uuid4

import httpx
import pytest

from noxus_sdk.client import Client
from noxus_sdk.resources.conversations import (
    ConversationFile,
    ConversationSettings,
    KnowledgeBaseSelectorTool,
    MessageRequest,
    NoxusQaTool,
    WebResearchTool,
)
from noxus_sdk.resources.knowledge_bases import KnowledgeBase


@pytest.fixture
def conversation_settings():
    return ConversationSettings(
        model=["gpt-4o"],
        temperature=0.7,
        tools=[NoxusQaTool()],
        max_tokens=1000,
        extra_instructions="You are a helpful assistant.",
    )


@pytest.mark.anyio
async def test_create_conversation(
    client: Client,
    conversation_settings: ConversationSettings,
):
    try:
        conversation = await client.conversations.acreate(
            name="Test Conversation",
            settings=conversation_settings,
        )
    except httpx.HTTPStatusError as e:
        print(e.response.text)
        raise e

    try:
        assert conversation.name == "Test Conversation"
        assert conversation.settings.model == ["gpt-4o"]
        assert conversation.settings.temperature == 0.7
        assert len(conversation.settings.tools) == 1

        # Test get conversation
        fetched = await client.conversations.aget(conversation.id)
        assert fetched.id == conversation.id
        assert fetched.name == conversation.name

    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
async def test_list_conversations(
    client: Client,
    conversation_settings: ConversationSettings,
):
    conv1 = await client.conversations.acreate(
        name="Test Conv 1",
        settings=conversation_settings,
    )
    conv2 = await client.conversations.acreate(
        name="Test Conv 2",
        settings=conversation_settings,
    )
    conversations = await client.conversations.alist()
    assert len(conversations) == 0

    await conv1.aadd_message(MessageRequest(content="Hello, world!"))
    await conv2.aadd_message(MessageRequest(content="Hello, world!"))

    try:
        conversations = await client.conversations.alist()
        assert len(conversations) == 2

        # Test pagination
        page1 = await client.conversations.alist()
        assert len(page1) == 2

    except Exception as e:
        print(e)
        raise e
    finally:
        await client.conversations.adelete(conv1.id)
        await client.conversations.adelete(conv2.id)


@pytest.mark.anyio
async def test_conversation_messages(
    client: Client,
    conversation_settings: ConversationSettings,
):
    conversation = await client.conversations.acreate(
        name="Test Messages",
        settings=conversation_settings,
    )

    try:
        # Add a message
        message = MessageRequest(content="Hello, world!")
        await conversation.aadd_message(message)

        # Get messages
        messages = await conversation.aget_messages()

        assert len(messages) >= 1
        assert any(
            any("Hello, world!" in part.get("content", "") for part in msg.message_parts) for msg in messages
        ), messages

    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
@pytest.mark.skip("yau")
async def test_conversation_with_kb(client: Client, kb: KnowledgeBase, test_file):
    import asyncio
    import time

    await kb.aupload_document([test_file], prefix="/test1")
    while kb.status not in ["training", "error"]:
        await kb.arefresh()
    assert kb.status in ["training"]

    timeout = time.time() + 120  # 60s timeout
    trained_docs = await kb.alist_documents(status="trained")
    while len(trained_docs) == 0 and timeout - time.time() > 0:
        trained_docs = await kb.alist_documents(status="trained")
        await asyncio.sleep(0.5)
    conversation_settings = ConversationSettings(
        model=["gpt-4o"],
        temperature=0.7,
        tools=[KnowledgeBaseSelectorTool()],
        max_tokens=1000,
    )

    conversation = await client.conversations.acreate(
        name="Test With KB",
        settings=conversation_settings,
    )

    try:
        message = MessageRequest(
            content="What is the capital of France?",
            kb_id=kb.id,
            tool="kb_qa",
        )
        await conversation.aadd_message(message)

        messages = await conversation.aget_messages()
        assert len(messages) >= 2
        assert any(
            # check if any message part has tool calls, meaning it called the kb
            any(part.get("role", None) == "function" for part in msg.message_parts)
            for msg in messages
        ), messages
    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
async def test_conversation_with_web_search(client: Client):
    conversation_settings = ConversationSettings(
        model=["gpt-4o"],
        temperature=0.7,
        tools=[WebResearchTool()],
        max_tokens=1000,
    )

    conversation = await client.conversations.acreate(
        name="Test With Web Search",
        settings=conversation_settings,
    )

    try:
        message = MessageRequest(
            content="What is the capital of France?",
            tool="web_research",
        )
        await conversation.aadd_message(message)

        messages = await conversation.aget_messages()
        assert len(messages) >= 2
        assert any(
            # check if any message part has tool calls, meaning it called the kb
            any(part.get("role", None) == "function" for part in msg.message_parts)
            for msg in messages
        ), messages
    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
async def test_conversation_with_noxus_qa(client: Client):
    conversation_settings = ConversationSettings(
        model=["gpt-4o"],
        temperature=0.7,
        tools=[NoxusQaTool()],
        max_tokens=1000,
    )

    conversation = await client.conversations.acreate(
        name="Test With Noxus QA",
        settings=conversation_settings,
    )

    try:
        message = MessageRequest(
            content="What is the capital of France?",
            tool="noxus_qa",
        )
        await conversation.aadd_message(message)

        messages = await conversation.aget_messages()
        assert len(messages) >= 2
        assert any(
            # check if any message part has tool calls, meaning it called the kb
            any(part.get("role", None) == "function" for part in msg.message_parts)
            for msg in messages
        ), messages
    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
async def test_conversation_with_file_b64(
    client: Client,
    conversation_settings: ConversationSettings,
):
    conversation = await client.conversations.acreate(
        name="Test With File",
        settings=conversation_settings,
    )

    try:
        file = ConversationFile(
            name="test.txt",
            status="success",
            b64_content=base64.b64encode(b"Hello, world!").decode("utf-8"),
        )

        message = MessageRequest(content="What does the file say?", files=[file])

        await conversation.aadd_message(message)
        messages = await conversation.aget_messages()
        assert len(messages) >= 1
        assert any("hello, world!" in part.get("content", "").lower() for msg in messages for part in msg.message_parts)

    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
async def test_update_conversation(
    client: Client,
    conversation_settings: ConversationSettings,
):
    conversation = await client.conversations.acreate(
        name="Original Name",
        settings=conversation_settings,
    )

    try:
        # Update settings
        new_settings = ConversationSettings(
            model=["gpt-3.5-turbo"],
            temperature=0.5,
            tools=[WebResearchTool()],
            max_tokens=1000,
        )

        updated = await client.conversations.aupdate(
            conversation.id,
            name="Updated Name",
            settings=new_settings,
        )

        assert updated.name == "Updated Name"
        assert updated.settings.model == ["gpt-3.5-turbo"]
        assert updated.settings.temperature == 0.5
        assert len(updated.settings.tools) == 1

    finally:
        await client.conversations.adelete(conversation.id)


@pytest.mark.anyio
async def test_create_nonexistant_with_agent(client: Client):
    agent_id = str(uuid4())  # Mock agent ID
    with pytest.raises(httpx.HTTPStatusError):
        conversation = await client.conversations.acreate(
            name="Agent Conversation",
            agent_id=agent_id,
        )


def test_invalid_creation_params(
    client: Client,
    conversation_settings: ConversationSettings,
):
    with pytest.raises(ValueError):  # noqa: PT011 - Legacy... TODO(Andre) - Improve exceptions and remove this
        client.conversations.create(
            name="Invalid",
            settings=conversation_settings,
            agent_id="some-id",
        )

    with pytest.raises(ValueError):  # noqa: PT011 - Legacy... TODO(Andre) - Improve exceptions and remove this
        client.conversations.create(name="Invalid")



"""Configuration display components."""

from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Literal

from pydantic import BaseModel, Field, model_validator
from pydantic import ConfigDict as PydanticConfigDict

# ========================================
# BASE CLASS
# ========================================


class BaseConfigDisplay(BaseModel):
    """Base class for configuration UI components.

    Provides common layout and styling properties. Subclasses should
    override the 'type' field with their component identifier.
    """

    type: str = "base"
    label: str
    tooltip: str | None = None
    error: str | None = None
    is_subconfig: bool = False
    col_span: int = 12  # Grid column span (1-12)
    col_justify: Literal["start", "center", "end"] | None = None
    col_align: Literal["start", "center", "end"] | None = None
    is_disabled: bool | None = False


# ========================================
# BASIC INPUT COMPONENTS
# ========================================


class ConfigText(BaseConfigDisplay):
    type: Literal["text"] = "text"  # type: ignore
    placeholder: str | None = None
    number_of_lines: int = 1
    show_refresh: bool = False
    is_code: bool = False
    is_textarea: bool = False
    class_name: str | None = None


class ConfigNumber(BaseConfigDisplay):
    type: Literal["number"] = "number"  # type: ignore


class ConfigNumberSlider(BaseConfigDisplay):
    type: Literal["number_slider"] = "number_slider"  # type: ignore
    min: float | int = 0
    max: float | int = 100
    step: float | int = 1
    hide_number: bool = True


class ConfigNumberRange(BaseConfigDisplay):
    type: Literal["number_range"] = "number_range"  # type: ignore
    min: float | int
    max: float | int


class ConfigToggle(BaseConfigDisplay):
    type: Literal["toggle"] = "toggle"  # type: ignore


class ConfigToggleButton(BaseConfigDisplay):
    type: Literal["toggle_button"] = "toggle_button"  # type: ignore
    icon: str
    toggledIcon: str | None = None  # noqa: N815 - Bad legacy code..


class ConfigFile(BaseConfigDisplay):
    type: Literal["file"] = "file"  # type: ignore


class ConfigFileArray(BaseConfigDisplay):
    type: Literal["file_array"] = "file_array"  # type: ignore
    can_select_multiple_files: bool = False
    allowed_file_types: list[str] | None = None


class ConfigTimePicker(BaseConfigDisplay):
    type: Literal["time_picker"] = "time_picker"  # type: ignore


class ConfigPhoneNumber(BaseConfigDisplay):
    type: Literal["phone_number"] = "phone_number"  # type: ignore


# ========================================
# SELECTION COMPONENTS
# ========================================


class ConfigSelect(BaseConfigDisplay):
    type: Literal["select"] = "select"  # type: ignore
    values: list[Any]
    placeholder: str | None = None
    is_clearable: bool = False
    is_horizontal: bool = False


class ConfigChipsSelect(BaseConfigDisplay):
    type: Literal["select_chips"] = "select_chips"  # type: ignore
    label: str | None = None  # type: ignore
    values: list[Any]
    can_wrap: bool = True


class ConfigMultiSelect(BaseConfigDisplay):
    type: Literal["multi_select"] = "multi_select"  # type: ignore
    values: list[Any]
    placeholder: str | None = None
    max_options: int | None = None


class ConfigTypeSelect(BaseConfigDisplay):
    type: Literal["select_type"] = "select_type"  # type: ignore
    values: list[Any]


class ConfigConditionSelect(BaseConfigDisplay):
    type: Literal["condition_select"] = "condition_select"  # type: ignore
    connector_name: str
    values: list[Any]


class ConfigEnumSlider(BaseConfigDisplay):
    type: Literal["enum_slider"] = "enum_slider"  # type: ignore
    values: list[str]
    is_horizontal: bool = False


class ConfigCoworkerSelect(BaseConfigDisplay):
    type: Literal["coworker_select"] = "coworker_select"  # type: ignore


class ConfigUserSelector(BaseConfigDisplay):
    type: Literal["user_selector"] = "user_selector"  # type: ignore
    label: str = ""
    placeholder: str | None = None


# ========================================
# MODEL & API COMPONENTS
# ========================================


class ConfigModelSelect(BaseConfigDisplay):
    type: Literal["model_select"] = "model_select"  # type: ignore
    endpoint: str
    model_type: str
    model_config = PydanticConfigDict(protected_namespaces=())


class ConfigModelToggle(BaseConfigDisplay):
    type: Literal["model_toggle"] = "model_toggle"  # type: ignore
    endpoint: str
    model_type: str
    model_config = PydanticConfigDict(protected_namespaces=())


class ConfigProviderModelToggle(BaseConfigDisplay):
    type: Literal["provider_model_toggle"] = "provider_model_toggle"  # type: ignore
    provider: str
    model: str
    icon: str | None = None
    icon_url: str | None = None

    @model_validator(mode="after")
    def _validate_icon(self) -> ConfigProviderModelToggle:
        if self.icon is None and self.icon_url is None:
            raise ValueError("Either icon or icon_url must be provided")
        return self


class ApiToggleState(str, Enum):
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    REQUESTING = "requesting"


class ConfigApiToggle(BaseConfigDisplay):
    type: Literal["api_toggle"] = "api_toggle"  # type: ignore
    toggle_label: str


class ConfigProviderApiToggle(ConfigApiToggle):
    type: Literal["provider_api_toggle"] = "provider_api_toggle"  # type: ignore


class ServiceAccountField(BaseConfigDisplay):
    type: Literal["service_account"] = "service_account"  # type: ignore
    name: str
    key: str


class APIKeyField(BaseConfigDisplay):
    type: Literal["api_key"] = "api_key"  # type: ignore
    name: str
    key: str


# ========================================
# TEXT & RICH TEXT COMPONENTS
# ========================================


class ConfigChipsText(BaseConfigDisplay):
    type: Literal["chips_text"] = "chips_text"  # type: ignore
    placeholder: str | None = None
    bottom_text: str | None = None
    chips_are_closable: bool = False


class ConfigChipsListText(BaseConfigDisplay):
    type: Literal["chips_list_text"] = "chips_list_text"  # type: ignore
    placeholder: str | None = None
    max_values: int | None = None
    bottom_text: str | None = None
    chips_are_closable: bool = False
    hide_toolbar: bool = False


class ConfigTextVariables(BaseConfigDisplay):
    type: Literal["big_text_variables"] = "big_text_variables"  # type: ignore
    placeholder: str | None = None
    number_of_lines: int = 1
    connector_name: str


class ConfigBigText(BaseConfigDisplay):
    type: Literal["big_text"] = "big_text"  # type: ignore
    placeholder: str | None = None
    number_of_lines: int = 5
    hide_toolbar: bool = False


class ConfigRichTextVariables(BaseConfigDisplay):
    type: Literal["rich_text_variables"] = "rich_text_variables"  # type: ignore
    placeholder: str | None = None
    number_of_lines: int = 5
    connector_name: str
    hide_toolbar: bool = False
    is_simple_text: bool = False
    no_space_on_insert: bool = False


class ConfigRichTextVariablesAI(BaseConfigDisplay):
    type: Literal["rich_text_variables_ai"] = "rich_text_variables_ai"  # type: ignore
    placeholder: str | None = None
    number_of_lines: int = 5
    connector_name: str

    @classmethod
    def render(cls, template: list[dict], variables: dict[str, str]) -> str:
        """Render template by replacing ((variable)) syntax with actual values."""
        text_template = "".join(m["text"] for m in template)
        for key, value in variables.items():
            text_template = text_template.replace(f"(({key}))", value)
        return text_template


class ConfigTextArray(BaseConfigDisplay):
    type: Literal["text_array"] = "text_array"  # type: ignore


class ConfigOutputSelector(BaseConfigDisplay):
    type: Literal["output_selector"] = "output_selector"  # type: ignore
    connector_name: str


class ConfigVariableInputValue(BaseConfigDisplay):
    type: Literal["variable_input_value"] = "variable_input_value"  # type: ignore
    types: list[str]
    hide_toolbar: bool = False


# ========================================
# DICTIONARY & COMPLEX DATA
# ========================================


class ConfigDisplayDict(BaseConfigDisplay):
    type: Literal["config_dict"] = "config_dict"  # type: ignore
    key_label: str
    value_label: str


class ConfigDictList(BaseConfigDisplay):
    type: Literal["config_dict_list"] = "config_dict_list"  # type: ignore
    keys: list[str]


class ConfigDictEntry(BaseModel):
    label: str
    placeholder: str
    has_chips: bool = False


class ConfigDictComplexList(BaseConfigDisplay):
    type: Literal["config_dict_complex_list"] = "config_dict_complex_list"  # type: ignore
    subtitle: str | None = None
    keys: list[ConfigDictEntry]
    prefix_name: str = "Output"


class ConfigDictComplexListStandalone(BaseConfigDisplay):
    type: Literal["config_dict_complex_list_standalone"] = "config_dict_complex_list_standalone"  # type: ignore
    subtitle: str | None = None
    keys: list[ConfigDictEntry]


class ConfigDictListWithoutConnector(BaseConfigDisplay):
    type: Literal["config_dict_list_without_connector"] = "config_dict_list_without_connector"  # type: ignore
    key_label: str
    value_label: str
    key_prefix: str
    value_placeholder: str
    required_keys: list[str] | None = None


class ConfigTypeDictArray(BaseConfigDisplay):
    type: Literal["type_dict_array"] = "type_dict_array"  # type: ignore
    values: list[str]
    type_map: dict[str, str]


# ========================================
# DYNAMIC & SEARCH COMPONENTS
# ========================================


class ConfigDynamicText(BaseConfigDisplay):
    type: Literal["config_dynamic_text"] = "config_dynamic_text"  # type: ignore
    placeholder: str | None = None


class ConfigDynamicSelect(BaseConfigDisplay):
    type: Literal["config_dynamic_select"] = "config_dynamic_select"  # type: ignore


class ConfigSearchBar(BaseConfigDisplay):
    type: Literal["search_bar"] = "search_bar"  # type: ignore
    search_endpoint: str


class ConfigWorkflowVersion(BaseConfigDisplay):
    type: Literal["config_workflow_version"] = "config_workflow_version"  # type: ignore


class ConfigComplexCondition(BaseConfigDisplay):
    type: Literal["config_complex_condition"] = "config_complex_condition"  # type: ignore


class ConfigScrapeWebsiteList(BaseConfigDisplay):
    type: Literal["config_scrape_website_list"] = "config_scrape_website_list"  # type: ignore


# ========================================
# THIRD-PARTY INTEGRATIONS
# ========================================


class GdriveFileType(str, Enum):
    PDFS = "application/pdf"
    DOCUMENTS = "application/pdf,text/plain,application/vnd.google-apps.document,application/vnd.google-apps.presentation,application/json"
    GDOCUMENTS = "application/vnd.google-apps.document"
    FOLDERS = "application/vnd.google-apps.folder"
    SPREADSHEETS = "application/vnd.google-apps.spreadsheet"
    ANY = "application/pdf,text/plain,application/vnd.google-apps.document,application/vnd.google-apps.presentation,application/json,application/vnd.google-apps.folder,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/vnd.openxmlformats-officedocument.presentationml.presentation"


class ConfigGdrivePicker(BaseConfigDisplay):
    model_config = PydanticConfigDict(use_enum_values=True)
    type: Literal["gdrive_picker"] = "gdrive_picker"  # type: ignore
    multi_select: bool = False
    file_type: GdriveFileType | None = None
    setIncludeFolders: bool = False  # noqa: N815 - Bad legacy code..


class ConfigOneDrivePicker(BaseConfigDisplay):
    model_config = PydanticConfigDict(use_enum_values=True)
    type: Literal["onedrive_picker"] = "onedrive_picker"  # type: ignore
    base_url: str | None = None
    multi_select: bool = False
    file_type: str | None = None
    set_include_folders: bool = False
    name: str = "OneDrive"


# ========================================
# TOOL & WORKFLOW COMPONENTS
# ========================================

ToolSections = Literal["QuickAction", "Flows", "Knowledge", "Nodes", "MCP", "Chatflow"]


class ConfigToolsSelect(BaseConfigDisplay):
    type: Literal["tools_select"] = "tools_select"  # type: ignore
    quick_tools: list[str] | None = None
    show_sections: list[ToolSections] = ["QuickAction", "Flows", "Knowledge"]


# ========================================
# UI DISPLAY COMPONENTS (NON-INPUT)
# ========================================


class ConfigDivider(BaseConfigDisplay):
    type: Literal["divider"] = "divider"  # type: ignore
    label: str = ""


class ConfigBanner(BaseConfigDisplay):
    type: Literal["banner"] = "banner"  # type: ignore
    label: str = ""
    style: Literal["info", "warning", "error"] = "info"


class ConfigTextDisplay(BaseConfigDisplay):
    type: Literal["ui_text_display"] = "ui_text_display"  # type: ignore
    label: str = ""
    title: str | None = None
    text: str | None = None
    small_text: str | None = None


AnyConfigDisplay = Annotated[
    ConfigText
    | ConfigNumber
    | ConfigNumberSlider
    | ConfigNumberRange
    | ConfigToggle
    | ConfigToggleButton
    | ConfigFile
    | ConfigFileArray
    | ConfigTimePicker
    | ConfigPhoneNumber
    | ConfigSelect
    | ConfigChipsSelect
    | ConfigMultiSelect
    | ConfigTypeSelect
    | ConfigConditionSelect
    | ConfigEnumSlider
    | ConfigCoworkerSelect
    | ConfigUserSelector
    | ConfigModelSelect
    | ConfigModelToggle
    | ConfigProviderModelToggle
    | ConfigApiToggle
    | ConfigChipsText
    | ConfigChipsListText
    | ConfigTextVariables
    | ConfigBigText
    | ConfigRichTextVariables
    | ConfigRichTextVariablesAI
    | ConfigTextArray
    | ConfigOutputSelector
    | ConfigVariableInputValue
    | ConfigDisplayDict
    | ConfigDictList
    | ConfigDictComplexList
    | ConfigDictComplexListStandalone
    | ConfigDictListWithoutConnector
    | ConfigTypeDictArray
    | ConfigDynamicText
    | ConfigDynamicSelect
    | ConfigSearchBar
    | ConfigWorkflowVersion
    | ConfigComplexCondition
    | ConfigScrapeWebsiteList
    | ConfigGdrivePicker
    | ConfigOneDrivePicker
    | ConfigToolsSelect
    | ConfigDivider
    | ConfigBanner
    | ConfigTextDisplay,
    Field(discriminator="type"),
]



from typing import TYPE_CHECKING

from noxus_sdk.workflows.workflow import WorkflowDefinition

if TYPE_CHECKING:
    from noxus_sdk.resources.conversations import Conversation


class AgentFlowDefinition(WorkflowDefinition):
    type: str = "agent_flow"

    def verify_name_legal(self, name: str) -> None:
        if name not in [
            "InputNode",
            "OutputNode",
            "FileInputNode",
            "ImageInputNode",
        ]:
            raise ValueError(f"Invalid node name: {name}")

    def update(self, *, force: bool = False) -> "AgentFlowDefinition":
        if not self.client:
            raise ValueError("Client not set")
        w = self.client.agentflows.update(self.id, self, force=force)
        self.refresh_from_data(client=self.client, **w.model_dump())
        return w

    async def aupdate(self, *, force: bool = False) -> "AgentFlowDefinition":
        if not self.client:
            raise ValueError("Client not set")
        w = await self.client.agentflows.aupdate(self.id, self, force=force)
        self.refresh_from_data(client=self.client, **w.model_dump())
        return w

    def save(self) -> "AgentFlowDefinition":
        if not self.client:
            raise ValueError("Client not set")
        return self.client.agentflows.save(self)

    async def asave(self) -> "AgentFlowDefinition":
        if not self.client:
            raise ValueError("Client not set")
        return await self.client.agentflows.asave(self)

    def run(self) -> "Conversation":  # type: ignore
        from noxus_sdk.resources.conversations import ConversationSettings

        if not self.client:
            raise ValueError("Client not set")
        return self.client.conversations.create(
            name=self.name,
            settings=ConversationSettings(
                agent_flow_id=self.id,
                model=[
                    "chat-balanced",
                    "gpt-4.1",
                    "claude-4-sonnet",
                    "claude-3.7-sonnet-thinking",
                    "claude-3.5-sonnet-v2",
                    "gemini-2.5-flash",
                    "gpt-4o",
                ],
                temperature=0.4,
                tools=[],
            ),
        )

    async def arun(self) -> "Conversation":  # type: ignore
        from noxus_sdk.resources.conversations import ConversationSettings

        if not self.client:
            raise ValueError("Client not set")
        return self.client.conversations.create(
            self.id,
            ConversationSettings(
                agent_flow_id=self.id,
                model=[
                    "chat-balanced",
                    "gpt-4.1",
                    "claude-4-sonnet",
                    "claude-3.7-sonnet-thinking",
                    "claude-3.5-sonnet-v2",
                    "gemini-2.5-flash",
                    "gpt-4o",
                ],
                temperature=0.4,
                tools=[],
            ),
        )



"""Plugin domain - everything related to plugin management"""

from noxus_sdk.plugins.base import BasePlugin, PluginConfiguration
from noxus_sdk.plugins.context import RemoteExecutionContext
from noxus_sdk.plugins.manifest import PluginManifest
from noxus_sdk.plugins.sources import (
    AnyPluginSource,
    GitPluginSource,
    LocalPluginSource,
    UploadPluginSource,
)
from noxus_sdk.plugins.types import PluginCategory, PluginSourceType, PluginStatus

__all__ = [
    "AnyPluginSource",
    "BasePlugin",
    "GitPluginSource",
    "LocalPluginSource",
    "PluginCategory",
    "PluginConfiguration",
    "PluginManifest",
    "PluginSourceType",
    "PluginStatus",
    "RemoteExecutionContext",
    "UploadPluginSource",
]
